# Copyright 2024 The Brax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Train."""
import copy
import datetime
import functools
from typing import Dict, Any
from brax.v1.experimental.braxlines import experiments
from brax.v1.experimental.braxlines.common import evaluators
from brax.v1.experimental.braxlines.common import logger_utils
from brax.v1.experimental.braxlines.envs.obs_indices import OBS_INDICES
from brax.v1.experimental.braxlines.irl_smm import evaluators as irl_evaluators
from brax.v1.experimental.braxlines.irl_smm import utils as irl_utils
from brax.v1.experimental.braxlines.training import ppo
from brax.v1.experimental.composer import composer
import jax

TASK_KEYS = (
    ('env_name',),
    ('obs_indices',),
    ('obs_scale',),
    ('target_num_modes',),
    ('target_num_samples',),
)


def train(train_job_params: Dict[str, Any],
          output_dir: str,
          return_dict: Dict[str, float] = None,
          progress_dict: Dict[str, float] = None,
          env_tag: str = None):
  """Train."""
  del env_tag
  return_dict = return_dict or {}
  progress_dict = progress_dict or {}
  logger_utils.save_config(
      f'{output_dir}/config.txt', train_job_params, verbose=True)
  config = copy.deepcopy(train_job_params)

  # @markdown **Task Parameters**
  env_name = config.pop('env_name', 'ant')
  obs_indices = config.pop('obs_indices', 'vel')
  obs_scale = config.pop('obs_scale', 5.0)
  obs_indices = OBS_INDICES[obs_indices][env_name]
  target_num_modes = config.pop('target_num_modes', 2)
  target_num_samples = config.pop('target_num_samples', 250)

  # @markdown **Experiment Parameters**
  reward_type = config.pop('reward_type', 'gail2')
  logits_clip_range = config.pop('logits_clip_range', 5.0)
  env_reward_multiplier = config.pop('env_reward_multiplier', 0.0)
  normalize_obs_for_disc = config.pop('normalize_obs_for_disc', False)
  balance_data_for_disc = config.pop('balance_data_for_disc', True)
  seed = config.pop('seed', 0)
  evaluate_dist = config.pop('evaluate_dist', False)
  spectral_norm = config.pop('spectral_norm', False)
  gradient_penalty_weight = config.pop('gradient_penalty_weight', 0.0)
  ppo_params = experiments.defaults.get_ppo_params(env_name, default='ant')
  ppo_params.update(config.pop('ppo_params', {}))
  assert not config, f'unused config: {config}'

  # generate target data
  rng = jax.random.PRNGKey(seed=seed)
  jit_get_dist = jax.jit(
      functools.partial(
          irl_utils.get_multimode_dist,
          indexed_obs_dim=len(obs_indices),
          num_modes=target_num_modes,
          scale=obs_scale))
  target_dist = jit_get_dist()
  target_data = target_dist.sample(seed=rng, sample_shape=(target_num_samples,))

  # make env_fn
  base_env_fn = composer.create_fn(env_name=env_name)
  base_env = base_env_fn()
  disc = irl_utils.IRLDiscriminator(
      input_size=len(obs_indices),
      obs_indices=obs_indices,
      obs_scale=obs_scale,
      include_action=False,
      arch=(32, 32),
      logits_clip_range=logits_clip_range,
      spectral_norm=spectral_norm,
      gradient_penalty_weight=gradient_penalty_weight,
      reward_type=reward_type,
      normalize_obs=normalize_obs_for_disc,
      balance_data=balance_data_for_disc,
      target_data=target_data,
      target_dist_fn=jit_get_dist,
      env=base_env)
  extra_params = disc.init_model(rng=jax.random.PRNGKey(seed=0))
  env_fn = irl_utils.create_fn(
      env_name=env_name,
      wrapper_params=dict(
          disc=disc,
          env_reward_multiplier=env_reward_multiplier,
      ))
  eval_env_fn = functools.partial(env_fn, auto_reset=False)
  # make inference functions and goals for evaluation
  core_env = env_fn()
  inference_fn = ppo.make_inference_fn(
      core_env.observation_size,
      core_env.action_size,
      normalize_observations=ppo_params.get('normalize_observation', True))
  inference_fn = jax.jit(inference_fn)

  tab = logger_utils.Tabulator(
      output_path=f'{output_dir}/training_curves.csv', append=False)
  # We determined some reasonable hyperparameters offline and share them here.
  train_fn = functools.partial(ppo.train, **ppo_params)

  times = [datetime.datetime.now()]
  plotpatterns = [
      'eval/episode_reward', 'losses/disc_loss', 'losses/total_loss',
      'losses/policy_loss', 'losses/value_loss', 'losses/entropy_loss',
      'metrics/energy_dist'
  ]

  def update_metrics_fn(num_steps, metrics, params):
    del num_steps
    if evaluate_dist:
      metrics.update(
          irl_evaluators.estimate_energy_distance_metric(
              params=params,
              disc=disc,
              target_data=target_data,
              env_fn=eval_env_fn,
              inference_fn=inference_fn,
              num_samples=10,
              time_subsampling=10,
              time_last_n=500,
              visualize=False,
              seed=0))

  progress, plot, _, _ = experiments.get_progress_fn(
      plotpatterns,
      times,
      tab=tab,
      max_ncols=5,
      return_dict=return_dict,
      progress_dict=progress_dict,
      xlim=[0, train_fn.keywords['num_timesteps']],
      update_metrics_fn=update_metrics_fn)

  extra_loss_fns = dict(disc_loss=disc.disc_loss_fn)
  inference_fn, params, _ = train_fn(
      environment_fn=env_fn,
      seed=seed,
      progress_fn=progress,
      extra_params=extra_params,
      extra_loss_fns=extra_loss_fns)
  plot(output_path=output_dir)
  irl_evaluators.visualize_disc(
      params=params, disc=disc, num_grid=25, output_path=output_dir)

  return_dict.update(dict(time_to_jit=times[1] - times[0]))
  return_dict.update(dict(time_to_train=times[-1] - times[1]))
  print(f'time to jit: {times[1] - times[0]}')
  print(f'time to train: {times[-1] - times[1]}')

  for i in range(3):
    evaluators.visualize_env(
        env_fn=eval_env_fn,
        inference_fn=inference_fn,
        params=params,
        batch_size=0,
        step_args=(params['normalizer'], params['extra']),
        output_path=output_dir,
        output_name=f'video_eps{i}',
    )

  metrics = irl_evaluators.estimate_energy_distance_metric(
      params=params,
      disc=disc,
      target_data=target_data,
      env_fn=eval_env_fn,
      inference_fn=inference_fn,
      num_samples=10,
      time_subsampling=10,
      time_last_n=500,
      visualize=True,
      output_path=output_dir,
      seed=0)
  return_dict.update(metrics)

  return return_dict



# Copyright 2024 The Brax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# pylint:disable=g-multiple-import
"""Some shared test functions."""
import copy
import time
from typing import Iterable, Tuple

from brax.base import System
from brax.io import mjcf
from etils import epath
import jax
from jax import numpy as jp
import mujoco
import numpy as np


def load_fixture_mujoco(path: str) -> mujoco.MjModel:
  full_path = epath.resource_path('brax') / f'test_data/{path}'
  if not full_path.exists():
    full_path = epath.resource_path('brax') / f'envs/assets/{path}'
  xml = mjcf.fuse_bodies(full_path.read_text())
  model = mujoco.MjModel.from_xml_string(xml)

  return model


def _normalize_q(model: mujoco.MjModel, q: np.ndarray):
  """Normalizes the quaternion part of q."""
  q = np.array(q)
  q_idx = 0
  for typ in model.jnt_type:
    q_dim = 7 if typ == 0 else 1
    if typ == 0:
      q[q_idx + 3:q_idx + 7] = (
          q[q_idx + 3:q_idx + 7] / np.linalg.norm(q[q_idx + 3:q_idx + 7]))
    q_idx += q_dim
  return q


def sample_mujoco_states(
    path: str, count: int = 500, modulo: int = 20, force_pgs: bool = False,
    random_init: bool = False, random_q_scale: float = 1.0,
    random_qd_scale: float = 0.1, vel_to_local: bool = True, seed: int = 42
) -> Iterable[Tuple[mujoco.MjData, mujoco.MjData]]:
  """Samples count / modulo states from mujoco for comparison."""
  np.random.seed(seed)
  model = load_fixture_mujoco(path)
  model.opt.iterations = 50  # return to default for high-precision comparison
  if force_pgs:
    model.opt.solver = 0
  data = mujoco.MjData(model)
  # give a little kick to avoid symmetry
  data.qvel = np.random.uniform(low=-0.01, high=0.01, size=(model.nv,))
  if random_init:
    data.qpos = np.random.uniform(model.nq) * random_q_scale
    data.qpos = _normalize_q(model, data.qpos)
    data.qvel = np.random.uniform(size=(model.nv,)) * random_qd_scale
  for i in range(count):
    before = copy.deepcopy(data)
    mujoco.mj_step(model, data)
    if i % modulo == 0:
      # hijack subtree_angmom, subtree_linvel (unused) to store xang, xvel
      for i in range(model.nbody):
        vel = np.zeros((6,))
        mujoco.mj_objectVelocity(
            model, data, mujoco.mjtObj.mjOBJ_XBODY.value, i, vel, vel_to_local)
        data.subtree_angmom[i] = vel[:3]
        data.subtree_linvel[i] = vel[3:]
      yield before, data


def load_fixture(path: str) -> System:
  full_path = epath.resource_path('brax') / f'test_data/{path}'
  if not full_path.exists():
    full_path = epath.resource_path('brax') / f'envs/assets/{path}'
  sys = mjcf.load(full_path)

  return sys


def benchmark(
    name: str, init_fn, step_fn, batch_size: int = 256, length: int = 1000
) -> float:
  """Reports jit time and op time for a function."""

  @jax.jit
  def run_batch(seed: jax.Array):
    rngs = jax.random.split(jax.random.PRNGKey(seed), batch_size)  # pytype: disable=wrong-arg-types  # jax-ndarray
    init_state = jax.vmap(init_fn)(rngs)

    @jax.vmap
    def run(state):
      def step(state, _):
        state = step_fn(state)
        return state, ()

      return jax.lax.scan(step, state, (), length=length)

    return run(init_state)

  times = []
  for i in range(5):
    t = time.time()
    jax.tree_util.tree_map(lambda x: x.block_until_ready(), run_batch(i))
    times.append(time.time() - t)
  op_time = jp.mean(jp.array(times[1:]))  # ignore JIT time

  batch_sps = batch_size * length / op_time
  print(
      f'{name} jit time: {times[0] - op_time:.3f}s, '
      f'op time: {op_time:.3f}s, '
      f'{batch_sps:,.0f} batch steps/sec on '
      f'{jax.devices()[0].device_kind}'
  )

  return batch_sps  # pytype: disable=bad-return-type  # jnp-type



# Copyright 2024 The Brax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Reward functions.

Each function has the form:
    reward, done = reward_fn(action, obs_dict, ...)
    If action.shape = (...,action_dim), reward/done shapes are (...,reward_dim)
"""
# pylint:disable=g-long-lambda
import functools
import inspect
from typing import Any, Dict, Union
from brax.v1.experimental.composer.observers import Observer
from brax.v1.experimental.composer.observers import SimObserver as so
from jax import lax
from jax import numpy as jnp


def get_default_kwargs(fn):
  spec = inspect.getfullargspec(fn)
  kwargs = {}
  if spec.defaults:
    kwargs.update(
        {k: v for k, v in zip(spec.args[-len(spec.defaults):], spec.defaults)})
  kwargs.update(spec.kwonlydefaults or {})
  return kwargs


def pop_wrapper_kwargs(reward_kwargs: Dict[str, Any]):
  wrapper_kwargs = get_default_kwargs(wrapper)
  wrapper_kwargs = {
      k: reward_kwargs.pop(k, v) for k, v in wrapper_kwargs.items()
  }
  return reward_kwargs, wrapper_kwargs


def wrapper(reward_fn,
            scale: float = 1.0,
            offset: float = 0.0,
            done_bonus: float = 0.0,
            exclude_from_score: bool = False):
  """Return both scale/offset reward and raw reward."""

  def fn(*args, **kwargs):
    reward, done = reward_fn(*args, **kwargs)
    if exclude_from_score:
      score = jnp.zeros_like(reward)
    else:
      score = reward
    reward = (reward + offset) * scale
    score *= jnp.sign(scale)
    reward = jnp.where(done, reward + done_bonus, reward)
    score = jnp.where(done, score + done_bonus, score)
    return reward, score, done

  return fn


def index_obs_dict(obs_dict: Dict[str, jnp.ndarray], obs: Union[Observer,
                                                                jnp.ndarray]):
  """Index obs_dict with observer."""
  if isinstance(obs, Observer):
    assert obs.name in obs_dict, f'{obs.name} not in {obs_dict.keys()}'
    obs = obs_dict[obs.name]
  return jnp.array(obs)


def constant_reward(action: jnp.ndarray,
                    obs_dict: Dict[str, jnp.ndarray],
                    value: float = 1.0):
  """Constant reward."""
  del obs_dict
  reward = jnp.ones(action.shape[:-1]) * value
  return reward, jnp.zeros_like(reward)


def control_reward(action: jnp.ndarray, obs_dict: Dict[str, jnp.ndarray]):
  """Negative Control reward."""
  del obs_dict
  ctrl_cost = jnp.linalg.norm(action, axis=-1)
  return -ctrl_cost, jnp.zeros_like(ctrl_cost)


def exp_norm_reward(action: jnp.ndarray, obs_dict: Dict[str, jnp.ndarray],
                    obs: Observer, **kwargs):
  """Exponential negative norm of an observation as reward."""
  reward, done = distance_reward(action, obs_dict, obs1=obs, obs2=0, **kwargs)
  return jnp.exp(reward), done


def direction_reward(action: jnp.ndarray,
                     obs_dict: Dict[str, jnp.ndarray],
                     vel0: Union[Observer, jnp.ndarray],
                     vel1: Union[Observer, jnp.ndarray],
                     pos0: Union[Observer, jnp.ndarray],
                     pos1: Union[Observer, jnp.ndarray],
                     sign: float = -1.0,
                     norm_kwargs: Dict[str, Any] = None):
  """Positive direction reward based on inner product."""
  del action
  norm_kwargs = norm_kwargs or {}
  vel0 = index_obs_dict(obs_dict, vel0)
  vel1 = index_obs_dict(obs_dict, vel1)
  pos0 = index_obs_dict(obs_dict, pos0)
  pos1 = index_obs_dict(obs_dict, pos1)
  ndim = max(vel0.ndim, vel1.ndim, pos0.ndim, pos1.ndim)
  vel0 = vel0.reshape((1,) * (ndim - vel0.ndim) + vel0.shape)
  vel1 = vel1.reshape((1,) * (ndim - vel1.ndim) + vel1.shape)
  pos0 = pos0.reshape((1,) * (ndim - pos0.ndim) + pos0.shape)
  pos1 = pos1.reshape((1,) * (ndim - pos1.ndim) + pos1.shape)
  agent_sign = jnp.sign(jnp.sum((pos1 - pos0) * vel0, axis=-1))
  opp_sign = jnp.sign(jnp.sum((pos0 - pos1) * vel1, axis=-1))
  # get unit vector & direction
  vel1 /= jnp.linalg.norm(vel1, axis=-1, **norm_kwargs)
  vel1 *= jnp.sign(sign)
  inner_product = lax.cond(
      agent_sign, lambda x: lax.cond(x, lambda y: jnp.sum(vel0 * y, axis=-1),
                                     lambda y: jnp.zeros_like(x), vel1),
      jnp.zeros_like, opp_sign)
  return jnp.clip(inner_product, a_min=0.0), jnp.zeros_like(inner_product)


def norm_reward(action: jnp.ndarray, obs_dict: Dict[str, jnp.ndarray],
                obs: Observer, **kwargs):
  """Negative norm of an observation as reward."""
  return distance_reward(action, obs_dict, obs1=obs, obs2=0, **kwargs)


def state_reward(action: jnp.ndarray, obs_dict: Dict[str, jnp.ndarray],
                 obs: Union[Observer, jnp.ndarray]):
  """State reward."""
  del action
  obs = index_obs_dict(obs_dict, obs)
  obs = jnp.sum(obs, axis=-1)
  done = jnp.zeros_like(obs)
  return obs, done


def distance_reward(action: jnp.ndarray,
                    obs_dict: Dict[str, jnp.ndarray],
                    obs1: Union[Observer, jnp.ndarray],
                    obs2: Union[Observer, jnp.ndarray],
                    max_dist: float = 1e8,
                    min_dist: float = 0,
                    norm_kwargs: Dict[str, Any] = None):
  """Negative distance reward."""
  del action
  norm_kwargs = norm_kwargs or {}
  obs1 = index_obs_dict(obs_dict, obs1)
  obs2 = index_obs_dict(obs_dict, obs2)
  ndim = max(obs1.ndim, obs2.ndim)
  obs1 = obs1.reshape((1,) * (ndim - obs1.ndim) + obs1.shape)
  obs2 = obs2.reshape((1,) * (ndim - obs2.ndim) + obs2.shape)
  delta = obs1 - obs2
  dist = jnp.linalg.norm(delta, axis=-1, **norm_kwargs)
  # instead of clipping, terminate
  # dist = jnp.clip(dist, a_min=min_dist, a_max=max_dist)
  done = jnp.zeros_like(dist)
  done = jnp.where(dist < min_dist, jnp.ones_like(done), done)
  done = jnp.where(dist > max_dist, jnp.ones_like(done), done)
  return -dist, done


def get_reward_fns(*components: Dict[str, Any],
                   reward_type: str = 'root_goal',
                   **reward_kwargs):
  """Get components-based reward functions.

  `reward_type` can be:
    - a string specifying common pre-defined reward functions, e.g. 'root_goal'
    - a callable reward_fn(actions, obs_dict, ...), in which case each default
      value in `reward_kwargs` can also be a function, in which case it will be
      processed as v=v(*components).

  Args:
    *components: component dictionaries from Composer().metadata.components
    reward_type: a str or a callable, specifying a reward function
    **reward_kwargs: kwargs arguments to the reward function

  Returns:
    a callable reward_fn(actions, obs_dict)
  """
  reward_kwargs, wrapper_kwargs = pop_wrapper_kwargs(reward_kwargs)
  if reward_type == 'root_goal':
    assert len(components) == 1, components
    sdtype = reward_kwargs.pop('sdtype', 'body')
    sdcomp = reward_kwargs.pop('sdcomp', 'pos')
    indices = reward_kwargs.pop('indices', None)
    target_goal = reward_kwargs.pop('target_goal')
    target_goal = jnp.array(target_goal)
    reward_fn = functools.partial(
        distance_reward,
        obs1=so(sdtype, sdcomp, components[0]['root'], indices=indices),
        obs2=target_goal,
        **reward_kwargs)
  elif reward_type == 'root_dist':
    assert len(components) == 2, components
    reward_fn = functools.partial(
        distance_reward,
        obs1=so('body', 'pos', components[0]['root']),
        obs2=so('body', 'pos', components[1]['root']),
        **reward_kwargs)
  elif callable(reward_type):
    reward_kwargs = {
        k: v(*components) if callable(v) else v
        for k, v in reward_kwargs.items()
    }
    reward_fn = functools.partial(reward_type, **reward_kwargs)
  else:
    raise NotImplementedError(reward_type)
  return wrapper(reward_fn, **wrapper_kwargs), reward_fn


def get_observers_from_reward_fns(reward_fn):
  """Get observers variable from reward_fn."""
  defaults = get_default_kwargs(reward_fn)
  return [v for _, v in sorted(defaults.items()) if isinstance(v, Observer)]



# Copyright 2024 The Brax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# pylint:disable=g-multiple-import
"""Exports a system config and trajectory as an html view."""

import base64
from typing import List, Optional, Union
import zlib

import brax
from brax.base import State, System
from brax.io import json
from etils import epath
import jinja2


def save(path: str, sys: System, states: List[State]):
  """Saves trajectory as an HTML text file."""
  path = epath.Path(path)
  if not path.parent.exists():
    path.parent.mkdir(parents=True)
  path.write_text(render(sys, states))


def render_from_json(
    sys: str, height: Union[int, str], colab: bool, base_url: Optional[str]
) -> str:
  """Returns an HTML string that visualizes the brax system json string."""
  html_path = epath.resource_path('brax') / 'visualizer/index.html'
  template = jinja2.Template(html_path.read_text())

  js_url = base_url
  if base_url is None:
    base_url = 'https://cdn.jsdelivr.net/gh/google/brax'
    js_url = f'{base_url}@v{brax.__version__}/brax/visualizer/js/viewer.js'

  sys = base64.b64encode(zlib.compress(bytes(sys, 'utf-8'))).decode('ascii')
  html = template.render(
      system_json_b64=sys, height=height, js_url=js_url, colab=colab
  )
  return html


def render(
    sys: System,
    states: List[State],
    height: Union[int, str] = 480,
    colab: bool = True,
    base_url: Optional[str] = None,
) -> str:
  """Returns an HTML string for the brax system and trajectory.

  Args:
    sys: brax System object
    states: list of system states to render
    height: the height of the render window
    colab: whether to use css styles for colab
    base_url: the base url for serving the visualizer files. By default, a CDN
      url is used

  Returns:
    string containing HTML for the brax visualizer
  """
  return render_from_json(json.dumps(sys, states), height, colab, base_url)



# Copyright 2024 The Brax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# pylint:disable=g-multiple-import, g-importing-member
"""Wrappers to support Brax training."""

from typing import Callable, Dict, Optional, Tuple

from brax.base import System
from brax.envs.base import Env, State, Wrapper
from flax import struct
import jax
from jax import numpy as jp


def wrap(
    env: Env,
    episode_length: int = 1000,
    action_repeat: int = 1,
    randomization_fn: Optional[
        Callable[[System], Tuple[System, System]]
    ] = None,
) -> Wrapper:
  """Common wrapper pattern for all training agents.

  Args:
    env: environment to be wrapped
    episode_length: length of episode
    action_repeat: how many repeated actions to take per step
    randomization_fn: randomization function that produces a vectorized system
      and in_axes to vmap over

  Returns:
    An environment that is wrapped with Episode and AutoReset wrappers.  If the
    environment did not already have batch dimensions, it is additional Vmap
    wrapped.
  """
  env = EpisodeWrapper(env, episode_length, action_repeat)
  if randomization_fn is None:
    env = VmapWrapper(env)
  else:
    env = DomainRandomizationVmapWrapper(env, randomization_fn)
  env = AutoResetWrapper(env)
  return env


class VmapWrapper(Wrapper):
  """Vectorizes Brax env."""

  def __init__(self, env: Env, batch_size: Optional[int] = None):
    super().__init__(env)
    self.batch_size = batch_size

  def reset(self, rng: jax.Array) -> State:
    if self.batch_size is not None:
      rng = jax.random.split(rng, self.batch_size)
    return jax.vmap(self.env.reset)(rng)

  def step(self, state: State, action: jax.Array) -> State:
    return jax.vmap(self.env.step)(state, action)


class EpisodeWrapper(Wrapper):
  """Maintains episode step count and sets done at episode end."""

  def __init__(self, env: Env, episode_length: int, action_repeat: int):
    super().__init__(env)
    self.episode_length = episode_length
    self.action_repeat = action_repeat

  def reset(self, rng: jax.Array) -> State:
    state = self.env.reset(rng)
    state.info['steps'] = jp.zeros(rng.shape[:-1])
    state.info['truncation'] = jp.zeros(rng.shape[:-1])
    return state

  def step(self, state: State, action: jax.Array) -> State:
    def f(state, _):
      nstate = self.env.step(state, action)
      return nstate, nstate.reward

    state, rewards = jax.lax.scan(f, state, (), self.action_repeat)
    state = state.replace(reward=jp.sum(rewards, axis=0))
    steps = state.info['steps'] + self.action_repeat
    one = jp.ones_like(state.done)
    zero = jp.zeros_like(state.done)
    episode_length = jp.array(self.episode_length, dtype=jp.int32)
    done = jp.where(steps >= episode_length, one, state.done)
    state.info['truncation'] = jp.where(
        steps >= episode_length, 1 - state.done, zero
    )
    state.info['steps'] = steps
    return state.replace(done=done)


class AutoResetWrapper(Wrapper):
  """Automatically resets Brax envs that are done."""

  def reset(self, rng: jax.Array) -> State:
    state = self.env.reset(rng)
    state.info['first_pipeline_state'] = state.pipeline_state
    state.info['first_obs'] = state.obs
    return state

  def step(self, state: State, action: jax.Array) -> State:
    if 'steps' in state.info:
      steps = state.info['steps']
      steps = jp.where(state.done, jp.zeros_like(steps), steps)
      state.info.update(steps=steps)
    state = state.replace(done=jp.zeros_like(state.done))
    state = self.env.step(state, action)

    def where_done(x, y):
      done = state.done
      if done.shape:
        done = jp.reshape(done, [x.shape[0]] + [1] * (len(x.shape) - 1))  # type: ignore
      return jp.where(done, x, y)

    pipeline_state = jax.tree.map(
        where_done, state.info['first_pipeline_state'], state.pipeline_state
    )
    obs = where_done(state.info['first_obs'], state.obs)
    return state.replace(pipeline_state=pipeline_state, obs=obs)


@struct.dataclass
class EvalMetrics:
  """Dataclass holding evaluation metrics for Brax.

  Attributes:
      episode_metrics: Aggregated episode metrics since the beginning of the
        episode.
      active_episodes: Boolean vector tracking which episodes are not done yet.
      episode_steps: Integer vector tracking the number of steps in the episode.
  """

  episode_metrics: Dict[str, jax.Array]
  active_episodes: jax.Array
  episode_steps: jax.Array


class EvalWrapper(Wrapper):
  """Brax env with eval metrics."""

  def reset(self, rng: jax.Array) -> State:
    reset_state = self.env.reset(rng)
    reset_state.metrics['reward'] = reset_state.reward
    eval_metrics = EvalMetrics(
        episode_metrics=jax.tree_util.tree_map(
            jp.zeros_like, reset_state.metrics
        ),
        active_episodes=jp.ones_like(reset_state.reward),
        episode_steps=jp.zeros_like(reset_state.reward),
    )
    reset_state.info['eval_metrics'] = eval_metrics
    return reset_state

  def step(self, state: State, action: jax.Array) -> State:
    state_metrics = state.info['eval_metrics']
    if not isinstance(state_metrics, EvalMetrics):
      raise ValueError(
          f'Incorrect type for state_metrics: {type(state_metrics)}'
      )
    del state.info['eval_metrics']
    nstate = self.env.step(state, action)
    nstate.metrics['reward'] = nstate.reward
    episode_steps = jp.where(
        state_metrics.active_episodes,
        nstate.info['steps'],
        state_metrics.episode_steps,
    )
    episode_metrics = jax.tree_util.tree_map(
        lambda a, b: a + b * state_metrics.active_episodes,
        state_metrics.episode_metrics,
        nstate.metrics,
    )
    active_episodes = state_metrics.active_episodes * (1 - nstate.done)

    eval_metrics = EvalMetrics(
        episode_metrics=episode_metrics,
        active_episodes=active_episodes,
        episode_steps=episode_steps,
    )
    nstate.info['eval_metrics'] = eval_metrics
    return nstate


class DomainRandomizationVmapWrapper(Wrapper):
  """Wrapper for domain randomization."""

  def __init__(
      self,
      env: Env,
      randomization_fn: Callable[[System], Tuple[System, System]],
  ):
    super().__init__(env)
    self._sys_v, self._in_axes = randomization_fn(self.sys)

  def _env_fn(self, sys: System) -> Env:
    env = self.env
    env.unwrapped.sys = sys
    return env

  def reset(self, rng: jax.Array) -> State:
    def reset(sys, rng):
      env = self._env_fn(sys=sys)
      return env.reset(rng)

    state = jax.vmap(reset, in_axes=[self._in_axes, 0])(self._sys_v, rng)
    return state

  def step(self, state: State, action: jax.Array) -> State:
    def step(sys, s, a):
      env = self._env_fn(sys=sys)
      return env.step(s, a)

    res = jax.vmap(step, in_axes=[self._in_axes, 0, 0])(
        self._sys_v, state, action
    )
    return res



# Copyright 2024 The Brax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Trains a robot arm to push a ball to a target."""

import brax.v1 as brax
from brax.v1 import jumpy as jp
from brax.v1.envs import env


class Pusher(env.Env):



  """
  ### Description

  "Pusher" is a multi-jointed robot arm which is very similar to that of a
  human.

  The goal is to move a target cylinder (called *object*) to a goal position
  using the robot's end effector (called *fingertip*). The robot consists of
  shoulder, elbow, forearm, and wrist joints.

  ### Action Space

  The action space is a `Box(-2, 2, (7,), float32)`. An action `(a, b)`
  represents the torques applied at the hinge joints.

  | Num | Action                                        | Control Min | Control Max | Name (in corresponding config) | Joint | Unit         |
  |-----|-----------------------------------------------|-------------|-------------|--------------------------------|-------|--------------|
  | 0   | Rotation of the panning the shoulder          | -1          | 1           | r_shoulder_pan_joint           | hinge | torque (N m) |
  | 1   | Rotation of the shoulder lifting joint        | -1          | 1           | r_shoulder_lift_joint          | hinge | torque (N m) |
  | 2   | Rotation of the shoulder rolling joint        | -1          | 1           | r_upper_arm_roll_joint         | hinge | torque (N m) |
  | 3   | Rotation of hinge joint that flexed the elbow | -1          | 1           | r_elbow_flex_joint             | hinge | torque (N m) |
  | 4   | Rotation of hinge that rolls the forearm      | -1          | 1           | r_forearm_roll_joint           | hinge | torque (N m) |
  | 5   | Rotation of flexing the wrist                 | -1          | 1           | r_wrist_flex_joint             | hinge | torque (N m) |
  | 6   | Rotation of rolling the wrist                 | -1          | 1           | r_wrist_roll_joint             | hinge | torque (N m) |

  ### Observation Space

  Observations consist of

  - Angle of rotational joints on the pusher
  - Angular velocities of rotational joints on the pusher
  - The coordinates of the fingertip of the pusher
  - The coordinates of the object to be moved
  - The coordinates of the goal position

  The observation is a `ndarray` with shape `(23,)` where the elements
  correspond to the table below. An analogy can be drawn to a human arm in order
  to help understand the state space, with the words flex and roll meaning the
  same as human joints.

  | Num | Observation                                              | Min  | Max | Name (in corresponding config) | Joint    | Unit                     |
  |-----|----------------------------------------------------------|------|-----|--------------------------------|----------|--------------------------|
  | 0   | Rotation of the panning the shoulder                     | -Inf | Inf | r_shoulder_pan_joint           | hinge    | angle (rad)              |
  | 1   | Rotation of the shoulder lifting joint                   | -Inf | Inf | r_shoulder_lift_joint          | hinge    | angle (rad)              |
  | 2   | Rotation of the shoulder rolling joint                   | -Inf | Inf | r_upper_arm_roll_joint         | hinge    | angle (rad)              |
  | 3   | Rotation of hinge joint that flexed the elbow            | -Inf | Inf | r_elbow_flex_joint             | hinge    | angle (rad)              |
  | 4   | Rotation of hinge that rolls the forearm                 | -Inf | Inf | r_forearm_roll_joint           | hinge    | angle (rad)              |
  | 5   | Rotation of flexing the wrist                            | -Inf | Inf | r_wrist_flex_joint             | hinge    | angle (rad)              |
  | 6   | Rotation of rolling the wrist                            | -Inf | Inf | r_wrist_roll_joint             | hinge    | angle (rad)              |
  | 7   | Rotational velocity of the panning the shoulder          | -Inf | Inf | r_shoulder_pan_joint           | hinge    | angular velocity (rad/s) |
  | 8   | Rotational velocity of the shoulder lifting joint        | -Inf | Inf | r_shoulder_lift_joint          | hinge    | angular velocity (rad/s) |
  | 9   | Rotational velocity of the shoulder rolling joint        | -Inf | Inf | r_upper_arm_roll_joint         | hinge    | angular velocity (rad/s) |
  | 10  | Rotational velocity of hinge joint that flexed the elbow | -Inf | Inf | r_elbow_flex_joint             | hinge    | angular velocity (rad/s) |
  | 11  | Rotational velocity of hinge that rolls the forearm      | -Inf | Inf | r_forearm_roll_joint           | hinge    | angular velocity (rad/s) |
  | 12  | Rotational velocity of flexing the wrist                 | -Inf | Inf | r_wrist_flex_joint             | hinge    | angular velocity (rad/s) |
  | 13  | Rotational velocity of rolling the wrist                 | -Inf | Inf | r_wrist_roll_joint             | hinge    | angular velocity (rad/s) |
  | 14  | x-coordinate of the fingertip of the pusher              | -Inf | Inf | tips_arm                       | slide    | position (m)             |
  | 15  | y-coordinate of the fingertip of the pusher              | -Inf | Inf | tips_arm                       | slide    | position (m)             |
  | 16  | z-coordinate of the fingertip of the pusher              | -Inf | Inf | tips_arm                       | slide    | position (m)             |
  | 17  | x-coordinate of the object to be moved                   | -Inf | Inf | object (obj_slidex)            | slide    | position (m)             |
  | 18  | y-coordinate of the object to be moved                   | -Inf | Inf | object (obj_slidey)            | slide    | position (m)             |
  | 19  | z-coordinate of the object to be moved                   | -Inf | Inf | object                         | cylinder | position (m)             |
  | 20  | x-coordinate of the goal position of the object          | -Inf | Inf | goal (goal_slidex)             | slide    | position (m)             |
  | 21  | y-coordinate of the goal position of the object          | -Inf | Inf | goal (goal_slidey)             | slide    | position (m)             |
  | 22  | z-coordinate of the goal position of the object          | -Inf | Inf | goal                           | sphere   | position (m)             |


  ### Rewards

  The reward consists of two parts:

  - *reward_near*: This reward is a measure of how far the *fingertip* of the
  pusher (the unattached end) is from the object, with a more negative value
  assigned for when the pusher's *fingertip* is further away from the target.
  It is calculated as the negative vector norm of (position of the fingertip -
  position of target), or *-norm("fingertip" - "target")*.
  - *reward_dist*: This reward is a measure of how far the object is from the
  target goal position, with a more negative value assigned for object is
  further away from the target. It is calculated as the negative vector norm of
  (position of the object - position of goal), or *-norm("object" - "target")*.
  - *reward_control*: A negative reward for penalising the pusher if it takes
  actions that are too large. It is measured as the negative squared Euclidean
  norm of the action, i.e. as *- sum(action<sup>2</sup>)*.

  Unlike other environments, Pusher does not allow you to specify weights for
  the individual reward terms. However, `info` does contain the keys
  *reward_dist* and *reward_ctrl*. Thus, if you'd like to weight the terms, you
  should create a wrapper that computes the weighted reward from `info`.

  ### Starting State

  All pusher (not including object and goal) states start in (0.0, 0.0, 0.0,
  0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0). A uniform noise in the
  range [-0.005, 0.005] is added to the velocity attributes only. The velocities
  of the object and goal are permanently set to 0. The object's x-position is
  selected uniformly between [-0.3, 0] while the y-position is selected
  uniformly between [-0.2, 0.2], and this process is repeated until the vector
  norm between the object's (x,y) position and origin is not greater than 0.17.
  The goal always have the same position of (0.45, -0.05, -0.323).

  The default *dt = 0.05*.

  ### Episode Termination

  The episode terminates when any of the following happens:

  1. The episode duration reaches a 1000 timesteps.

  ### Arguments

  No additional arguments are currently supported (in v2 and lower), but
  modifications can be made to the XML file in the assets folder (or by changing
  the path to a modified XML file in another folder)..

  ```
  env = gym.make('Pusher-v2')
  ```

  There is no v3 for Pusher, unlike the robot environments where a v3 and
  beyond take gym.make kwargs such as xml_file, ctrl_cost_weight, reset_noise_scale etc.

  There is a v4 version that uses the mujoco-bindings
  ```
  env = gym.make('Pusher-v4')
  ```

  And a v5 version that uses Brax:

  ```
  env = gym.make('Pusher-v5')
  ```

  ### Version History

  * v5: ported to Brax.
  * v4: all mujoco environments now use the mujoco bindings in mujoco>=2.1.3
  * v2: All continuous control environments now use mujoco_py >= 1.50
  * v1: max_time_steps raised to 1000 for robot based tasks (not including
    reacher, which has a max_time_steps of 50). Added reward_threshold to
    environments.
  * v0: Initial versions release (1.0.0)
  """


  def __init__(self, legacy_spring=False, **kwargs):
    del legacy_spring
    super().__init__(_SYSTEM_CONFIG, **kwargs)
    self._object_idx = self.sys.body.index['object']
    self._tips_arm_idx = self.sys.body.index['r_wrist_roll_link']
    self._goal_idx = self.sys.body.index['goal']
    self._table_idx = self.sys.body.index['table']
    self._goal_pos = jp.array([0.45, 0.05, 0.05])

  def reset(self, rng: jp.ndarray) -> env.State:
    rng, rng1, rng2 = jp.random_split(rng, 3)

    # randomly orient object
    cylinder_pos = jp.concatenate([
        jp.random_uniform(rng, (1,), -0.3, 0),
        jp.random_uniform(rng1, (1,), -0.2, 0.2),
        jp.ones(1) * 0.0
    ])

    # constraint maximum distance of object
    norm = jp.norm(cylinder_pos)
    scale = jp.where(norm > .17, .17 / norm, 1.)  # pytype: disable=wrong-arg-types  # jax-ndarray
    cylinder_pos = scale * cylinder_pos + jp.array([0., 0., .05])
    qpos = self.sys.default_angle()

    qvel = jp.concatenate([
        jp.random_uniform(rng2, (self.sys.num_joint_dof - 4,), -0.005, 0.005),
        jp.zeros(4)
    ])
    qp = self.sys.default_qp(joint_angle=qpos, joint_velocity=qvel)

    # position cylinder and goal
    pos = jp.index_update(qp.pos, self._goal_idx, self._goal_pos)
    pos = jp.index_update(pos, self._object_idx, cylinder_pos)
    pos = jp.index_update(pos, self._table_idx, jp.zeros(3))

    qp = qp.replace(pos=pos)

    obs = self._get_obs(qp)
    reward, done, zero = jp.zeros(3)
    metrics = {'reward_dist': zero, 'reward_ctrl': zero, 'reward_near': zero}
    return env.State(qp, obs, reward, done, metrics)

  def step(self, state: env.State, action: jp.ndarray) -> env.State:
    vec_1 = state.qp.pos[self._object_idx] - state.qp.pos[self._tips_arm_idx]
    vec_2 = state.qp.pos[self._object_idx] - state.qp.pos[self._goal_idx]

    reward_near = -jp.norm(vec_1)
    reward_dist = -jp.norm(vec_2)
    reward_ctrl = -jp.square(action).sum()

    qp, _ = self.sys.step(state.qp, action)
    obs = self._get_obs(qp)
    reward = reward_dist + 0.1 * reward_ctrl + 0.5 * reward_near
    state.metrics.update(
        reward_near=reward_near,
        reward_dist=reward_dist,
        reward_ctrl=reward_ctrl,
    )
    return state.replace(qp=qp, obs=obs, reward=reward)

  def _get_obs(self, qp: brax.QP) -> jp.ndarray:
    """Observe pusher body position and velocities."""
    joint_angle, joint_vel = self.sys.joints[0].angle_vel(qp)

    return jp.concatenate([
        joint_angle,
        joint_vel,
        qp.pos[self._tips_arm_idx],
        qp.pos[self._object_idx],
        qp.pos[self._goal_idx],
    ])


_SYSTEM_CONFIG = """
bodies {
  name: "table"
  colliders {
    plane {
    }
  }
  inertia {
    x: 1.0
    y: 1.0
    z: 1.0
  }
  mass: 1.0
  frozen {
    position {
      x: 1.0
      y: 1.0
      z: 1.0
    }
    rotation {
      x: 1.0
      y: 1.0
      z: 1.0
    }
  }
}
bodies {
  name: "r_shoulder_pan_link"
  colliders {
    position {
      z: -0.10000000149011612
    }
    rotation {
      y: -0.0
    }
    capsule {
      radius: 0.10000000149011612
      length: 0.800000011920929
    }
  }
  inertia {
    x: 1.0
    y: 1.0
    z: 1.0
  }
  mass: 1.0
}
bodies {
  name: "r_shoulder_lift_link"
  colliders {
    position {
    }
    rotation {
      x: -90.0
    }
    capsule {
      radius: 0.10000000149011612
      length: 0.4000000059604645
    }
  }
  inertia {
    x: 1.0
    y: 1.0
    z: 1.0
  }
  mass: 1.0
}
bodies {
  name: "r_upper_arm_roll_link"
  colliders {
    position {
    }
    rotation {
      x: -0.0
      y: 90.0
    }
    capsule {
      radius: 0.019999999552965164
      length: 0.23999999463558197
    }
  }
  colliders {
    position {
      x: 0.20000000298023224
    }
    rotation {
      x: -0.0
      y: 90.0
    }
    capsule {
      radius: 0.05999999865889549
      length: 0.5199999809265137
    }
  }
  inertia {
    x: 1.0
    y: 1.0
    z: 1.0
  }
  mass: 1.0
}
bodies {
  name: "r_elbow_flex_link"
  colliders {
    position {
    }
    rotation {
      x: -90.0
    }
    capsule {
      radius: 0.05999999865889549
      length: 0.1599999964237213
    }
  }
  inertia {
    x: 1.0
    y: 1.0
    z: 1.0
  }
  mass: 1.0
}
bodies {
  name: "r_forearm_roll_link"
  colliders {
    position {
    }
    rotation {
      x: -0.0
      y: 90.0
    }
    capsule {
      radius: 0.019999999552965164
      length: 0.23999999463558197
    }
  }
  colliders {
    position {
      x: 0.14550000429153442
    }
    rotation {
      x: -0.0
      y: 90.0
    }
    capsule {
      radius: 0.05000000074505806
      length: 0.39100000262260437
    }
  }
  inertia {
    x: 1.0
    y: 1.0
    z: 1.0
  }
  mass: 1.0
}
bodies {
  name: "r_wrist_flex_link"
  colliders {
    position {
    }
    rotation {
      x: -90.0
    }
    capsule {
      radius: 0.009999999776482582
      length: 0.05999999865889549
    }
  }
  inertia {
    x: 1.0
    y: 1.0
    z: 1.0
  }
  mass: 1.0
}
bodies {
  name: "r_wrist_roll_link"
  colliders {
    position {
    }
    rotation {
      x: -90.0
    }
    capsule {
      radius: 0.019999999552965164
      length: 0.23999999463558197
    }
  }
  colliders {
    position {
      x: 0.05000000074505806
      y: -0.10000000149011612
    }
    rotation {
      x: -0.0
      y: 90.0
    }
    capsule {
      radius: 0.019999999552965164
      length: 0.14000000059604645
    }
  }
  colliders {
    position {
      x: 0.05000000074505806
      y: 0.10000000149011612
    }
    rotation {
      x: -0.0
      y: 90.0
    }
    capsule {
      radius: 0.019999999552965164
      length: 0.14000000059604645
    }
  }
  colliders {
    position {
      x: 0.10000000149011612
      y: -0.10000000149011612
    }
    sphere {
      radius: 0.009999999776482582
    }
  }
  colliders {
    position {
      x: 0.10000000149011612
      y: 0.10000000149011612
    }
    sphere {
      radius: 0.009999999776482582
    }
  }
  inertia {
    x: 1.0
    y: 1.0
    z: 1.0
  }
  mass: 1.0
}
bodies {
  name: "object"
  colliders {
    position {
    }
    sphere {
      radius: 0.05000000074505806
    }
  }
  inertia {
    x: 1.0
    y: 1.0
    z: 1.0
  }
  mass: 1.0
  frozen {
    position { z: 1 }
    rotation { x: 1 y: 1 z: 1 }
  }
}
bodies {
  name: "goal"
  colliders {
    position {
    }
    rotation {
      y: -0.0
    }
    capsule {
      radius: 0.07999999821186066
      length: 0.16
    }
    color: "red"
  }
  inertia {
    x: 1.0
    y: 1.0
    z: 1.0
  }
  mass: 1.0
  frozen {
    all: true
  }
}
bodies {
  name: "mount"
  inertia {
    x: 1.0
    y: 1.0
    z: 1.0
  }
  mass: 1.0
  frozen {
    all: true
  }
}
joints {
  name: "r_shoulder_pan_joint"
  parent: "mount"
  child: "r_shoulder_pan_link"
  parent_offset {
    y: -0.6000000238418579
  }
  child_offset {
  }
  rotation {
    y: -90.0
  }
  angular_damping: 20.0
  angle_limit {
    min: -130.9437713623047
    max: 98.23945617675781
  }
  reference_rotation {
    y: -0.0
  }
}
joints {
  name: "r_shoulder_lift_joint"
  parent: "r_shoulder_pan_link"
  child: "r_shoulder_lift_link"
  parent_offset {
    x: 0.10000000149011612
  }
  child_offset {
  }
  rotation {
    y: -0.0
    z: 90.0
  }
  angular_damping: 20.0
  angle_limit {
    min: -30.00006866455078
    max: 80.0020980834961
  }
  reference_rotation {
    y: -0.0
  }
}
joints {
  name: "r_upper_arm_roll_joint"
  parent: "r_shoulder_lift_link"
  child: "r_upper_arm_roll_link"
  parent_offset {
  }
  child_offset {
  }
  rotation {
    y: -0.0
  }
  angular_damping: 20.0
  angle_limit {
    min: -85.94367218017578
    max: 97.40282440185547
  }
  reference_rotation {
    y: -0.0
  }
}
joints {
  name: "r_elbow_flex_joint"
  parent: "r_upper_arm_roll_link"
  child: "r_elbow_flex_link"
  parent_offset {
    x: 0.4000000059604645
  }
  child_offset {
  }
  rotation {
    y: -0.0
    z: 90.0
  }
  angular_damping: 20.0
  angle_limit {
    min: -133.00070190429688
  }
  reference_rotation {
    y: -0.0
  }
}
joints {
  name: "r_forearm_roll_joint"
  parent: "r_elbow_flex_link"
  child: "r_forearm_roll_link"
  parent_offset {
  }
  child_offset {
  }
  rotation {
    y: -0.0
  }
  angular_damping: 20.0
  angle_limit {
    min: -85.94367218017578
    max: 85.94367218017578
  }
  reference_rotation {
    y: -0.0
  }
}
joints {
  name: "r_wrist_flex_joint"
  parent: "r_forearm_roll_link"
  child: "r_wrist_flex_link"
  parent_offset {
    x: 0.32100000977516174
  }
  child_offset {
  }
  rotation {
    y: -0.0
    z: 90.0
  }
  angular_damping: 20.0
  angle_limit {
    min: -62.681583404541016
  }
  reference_rotation {
    y: -0.0
  }
}
joints {
  name: "r_wrist_roll_joint"
  parent: "r_wrist_flex_link"
  child: "r_wrist_roll_link"
  parent_offset {
  }
  child_offset {
  }
  rotation {
    y: -0.0
  }
  angular_damping: 20.0
  angle_limit {
    min: -85.94367218017578
    max: 85.94367218017578
  }
  reference_rotation {
    y: -0.0
  }
}
actuators {
  name: "r_shoulder_pan_joint"
  joint: "r_shoulder_pan_joint"
  strength: 100.0
  torque {
  }
}
actuators {
  name: "r_shoulder_lift_joint"
  joint: "r_shoulder_lift_joint"
  strength: 100.0
  torque {
  }
}
actuators {
  name: "r_upper_arm_roll_joint"
  joint: "r_upper_arm_roll_joint"
  strength: 100.0
  torque {
  }
}
actuators {
  name: "r_elbow_flex_joint"
  joint: "r_elbow_flex_joint"
  strength: 100.0
  torque {
  }
}
actuators {
  name: "r_forearm_roll_joint"
  joint: "r_forearm_roll_joint"
  strength: 100.0
  torque {
  }
}
actuators {
  name: "r_wrist_flex_joint"
  joint: "r_wrist_flex_joint"
  strength: 100.0
  torque {
  }
}
actuators {
  name: "r_wrist_roll_joint"
  joint: "r_wrist_roll_joint"
  strength: 100.0
  torque {
  }
}
gravity {  # zero-gravity works in this env because the `object` body is constrained to the x-y plane via `frozen` fields
}
collide_include {
  first: "table"
  second: "object"
}
collide_include {
  first: "r_wrist_roll_link"
  second: "object"
}
collide_include {
  first: "r_forearm_roll_link"
  second: "object"
}
collide_include {
  first: "r_forearm_roll_link"
  second: "table"
}
collide_include {
  first: "r_wrist_roll_link"
  second: "table"
}
dt: 0.05000000074505806
substeps: 50
solver_scale_pos: 0.20000000298023224
solver_scale_ang: 0.20000000298023224
solver_scale_collide: 0.20000000298023224
dynamics_mode: "pbd",
"""



# Copyright 2024 The Brax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Function to resolve collisions."""
# pylint:disable=g-multiple-import
from brax import contact
from brax import math
from brax.base import Force, Motion, System, Transform
from brax.spring.base import State
import jax
from jax import numpy as jp
from jax.ops import segment_sum


def resolve(sys: System, state: State) -> Motion:
  """Resolves springy collision constraint.

  Args:
    sys: System to forward propagate
    state: spring pipeline state

  Returns:
    xdv_i: delta-velocity to apply to link center of mass in world frame
  """
  c = contact.get(sys, state.x)

  if c is None:
    return Motion.zero((sys.num_links(),))

  @jax.vmap
  def impulse(c, link_idx, x_i, xd_i, i_inv, i_mass):
    """Calculates a velocity level update with baumgarte stabilization."""
    rel_pos = c.pos - x_i.pos
    rel_vel = xd_i.vel + jax.vmap(jp.cross)(xd_i.ang, rel_pos)
    rel_vel *= (link_idx > -1).reshape(-1, 1)
    contact_vel = rel_vel[0] - rel_vel[1]
    normal_vel = jp.dot(-c.frame[0], contact_vel)

    # the form of these update equations closely follows the smooth constraint
    # solver in Heiden et al's Tiny Differentiable Simulator
    temp = jax.vmap(jp.dot)(i_inv, jp.cross(rel_pos, -c.frame[0]))
    ang = jp.dot(-c.frame[0], jp.sum(jp.cross(temp, rel_pos), axis=0))
    baumgarte_vel = sys.baumgarte_erp / sys.opt.timestep * c.dist
    impulse = (-1.0 * (1.0 + c.elasticity) * normal_vel - baumgarte_vel) / (
        i_mass[0] + i_mass[1] + ang
    )
    impulse_vec = impulse * -c.frame[0]

    # apply drag due to friction acting parallel to the surface contact
    vel_d = contact_vel + normal_vel * c.frame[0]
    dir_d = vel_d / (1e-6 + math.safe_norm(vel_d))
    temp = jax.vmap(jp.dot)(i_inv, jp.cross(rel_pos, dir_d))
    ang_d = jp.dot(dir_d, jp.sum(jp.cross(temp, rel_pos), axis=0))
    impulse_d = math.safe_norm(vel_d) / (i_mass[0] + i_mass[1] + ang_d)

    # drag magnitude cannot exceed max friction
    impulse_d = jp.minimum(impulse_d, c.friction[0] * impulse)
    impulse_d_vec = -1.0 * impulse_d * dir_d

    # apply collision if penetrating, approaching, and oriented correctly
    apply_n = (c.dist < 0.0) & (normal_vel < 0) & (impulse > 0.0)
    # apply drag if moving laterally above threshold
    apply_d = apply_n * (math.safe_norm(vel_d) > 1e-3)

    f = Force.create(vel=impulse_vec * apply_n + impulse_d_vec * apply_d)

    return f, jp.array(apply_n, dtype=jp.float32)

  link_idx = jp.array(c.link_idx).T
  x_i, xd_i = state.x_i.take(link_idx), state.xd_i.take(link_idx)
  i_inv = state.i_inv.take(link_idx) * (link_idx > -1)
  i_mass = 1 / state.mass.take(link_idx) * (link_idx > -1)
  p, is_contact = impulse(c, link_idx, x_i, xd_i, i_inv, i_mass)

  # calculate the impulse to each link center of mass
  p = jax.tree.map(lambda x: jp.concatenate((x, -x)), p)
  pos = jp.tile(c.pos, (2, 1))
  link_idx = jp.concatenate(c.link_idx)
  xp_i = Transform.create(pos=pos - state.x_i.take(link_idx).pos).vmap().do(p)
  xp_i = jax.tree.map(lambda x: segment_sum(x, link_idx, sys.num_links()), xp_i)

  # average the impulse across multiple contacts
  num_contacts = segment_sum(jp.tile(is_contact, 2), link_idx, sys.num_links())
  xp_i = xp_i / (num_contacts.reshape((-1, 1)) + 1e-8)

  # convert impulse to delta-velocity
  xdv_i = Motion(
      vel=jax.vmap(lambda x, y: x / y)(xp_i.vel, state.mass),
      ang=jax.vmap(lambda x, y: x @ y)(state.i_inv, xp_i.ang),
  )

  return xdv_i



# Copyright 2024 The Brax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Functions to resolve collisions."""
# pylint:disable=g-multiple-import
from typing import Optional, Tuple

from brax import com
from brax import math
from brax.base import Contact, Motion, System, Transform, Force
from brax.positional.base import State
import jax
from jax import numpy as jp
from jax.ops import segment_sum


def resolve_position(
    sys: System,
    state: State,
    x_i_prev: Transform,
    contact: Optional[Contact],
) -> Tuple[Transform, jax.Array]:
  """Resolves positional collision constraint.

  The update equations follow section 3.5 of Müller et al.'s Extended Position
  Based Dynamics, where we have removed the compliance terms.
  (Müller, Matthias, et al. "Detailed rigid body simulation with extended
  position based dynamics." Computer Graphics Forum. Vol. 39. No. 8. 2020.).

  Args:
    sys: System to forward propagate
    state: positional pipeline state
    x_i_prev: center of mass position from previous step
    contact: Contact pytree

  Returns:
    x_i: new position after update
    dlambda: normal force information
  """
  if contact is None:
    return state.x_i, jp.zeros((1,))

  inv_mass = 1 / (sys.link.inertia.mass ** (1 - sys.spring_mass_scale))
  inv_inertia = com.inv_inertia(sys, state.x)

  @jax.vmap
  def translate(contact):
    link_idx = jp.array(contact.link_idx)
    x = state.x_i.concatenate(Transform.zero((1,))).take(link_idx)
    x_prev = x_i_prev.concatenate(Transform.zero((1,))).take(link_idx)

    # TODO: rewrite these updates to use pbd methods
    n = -contact.frame[0]
    c = contact.dist
    pos_p = contact.pos + n * c / 2.0 - x.pos[0]
    pos_c = contact.pos - n * c / 2.0 - x.pos[1]

    i_inv = inv_inertia.take(link_idx, axis=0)
    i_inv *= (link_idx > -1).reshape(-1, 1, 1)
    mass_inv = inv_mass.take(link_idx) * (link_idx > -1)

    # only spherical inertia effects
    cr1, cr2 = jp.cross(pos_p, n), jp.cross(pos_c, n)
    w1 = mass_inv[0] + jp.dot(cr1, i_inv[0] @ cr1)
    w2 = mass_inv[1] + jp.dot(cr2, i_inv[1] @ cr2)

    dlambda = -c / (w1 + w2 + 1e-6)
    coll_mask = c < 0
    p = dlambda * n * coll_mask

    dp_p_pos, dp_c_pos = p * mass_inv[0], -p * mass_inv[1]
    dp_p_rot = math.vec_quat_mul(i_inv[0] @ jp.cross(pos_p, p), x.rot[0])
    dp_c_rot = -math.vec_quat_mul(i_inv[1] @ jp.cross(pos_c, p), x.rot[1])

    # static friction
    q1inv, q2inv = jax.vmap(math.quat_inv)(x.rot)
    r1 = math.rotate(contact.pos - x.pos[0], q1inv)
    r2 = math.rotate(contact.pos - x.pos[1], q2inv)
    p1bar = x_prev.pos[0] + math.rotate(r1, x_prev.rot[0])
    p2bar = x_prev.pos[1] + math.rotate(r2, x_prev.rot[1])
    p0 = contact.pos

    deltap = (p0 - p1bar) - (p0 - p2bar)
    deltap_t = deltap - jp.dot(deltap, n) * n

    pos_p, pos_c = contact.pos - x.pos
    c = math.safe_norm(deltap_t)
    n = deltap_t / (c + 1e-6)

    cr1, cr2 = jp.cross(pos_p, n), jp.cross(pos_c, n)
    w1 = mass_inv[0] + jp.dot(cr1, i_inv[0] @ cr1)
    w2 = mass_inv[1] + jp.dot(cr2, i_inv[1] @ cr2)

    dlambdat = -c / (w1 + w2)
    static_mask = jp.where(jp.abs(dlambdat) < jp.abs(dlambda), 1.0, 0.0)
    p = dlambdat * n * static_mask * coll_mask

    dp_p_pos += p * mass_inv[0]
    dp_p_rot += 0.5 * math.vec_quat_mul(i_inv[0] @ jp.cross(pos_p, p), x.rot[0])
    dp_p = Transform(pos=dp_p_pos, rot=dp_p_rot) * sys.collide_scale

    dp_c_pos -= p * mass_inv[1]
    dp_c_rot -= 0.5 * math.vec_quat_mul(i_inv[1] @ jp.cross(pos_c, p), x.rot[1])
    dp_c = Transform(pos=dp_c_pos, rot=dp_c_rot) * sys.collide_scale

    return dp_p, dp_c, dlambda * coll_mask

  dp_p, dp_c, dlambda = translate(contact)
  dp = jax.tree.map(lambda x, y: jp.vstack([x, y]), dp_p, dp_c)
  dp = jax.tree.map(lambda x: jp.where(jp.isnan(x), 0.0, x), dp)
  link_idx = jp.concatenate(contact.link_idx)
  dp *= link_idx.reshape((-1, 1)) > -1
  dp = jax.tree.map(
      lambda f: jax.ops.segment_sum(f, link_idx, sys.num_links()), dp
  )
  x_i = state.x_i + dp
  x_i = x_i.replace(rot=jax.vmap(math.normalize)(x_i.rot)[0])

  return x_i, dlambda


def resolve_velocity(
    sys: System,
    state: State,
    xd_i_prev: Motion,
    contact: Contact,
    dlambda: jax.Array,
) -> Motion:
  """Velocity-level collision update for position based dynamics.

  The update equations here follow section 3.6 of Müller et al.'s Extended
  Position Based Dynamics. (Müller, Matthias, et al. "Detailed rigid body
  simulation with extended position based dynamics." Computer Graphics Forum.
  Vol. 39. No. 8. 2020.).

  Args:
    sys: System to update
    state: positional pipeline state
    xd_i_prev: velocity immediately preceding PBD velocity projection
    contact: Contact information for collision
    dlambda: Normal force of contact times time step squared

  Returns:
    Velocity level update for system state.
  """
  if contact is None:
    return Motion.zero((sys.num_links(),))

  x_i = state.x_i.concatenate(Transform.zero((1,)))
  inv_mass = 1 / (sys.link.inertia.mass ** (1 - sys.spring_mass_scale))
  inv_inertia = com.inv_inertia(sys, state.x)

  @jax.vmap
  def impulse(contact, dlambda):
    link_idx = jp.array(contact.link_idx)
    x = x_i.take(link_idx)
    xd = state.xd_i.concatenate(Motion.zero((1,))).take(link_idx)
    xd_prev = xd_i_prev.concatenate(Motion.zero((1,))).take(link_idx)
    i_inv = inv_inertia.take(link_idx, axis=0)
    i_inv *= (link_idx > -1).reshape(-1, 1, 1)
    mass_inv = inv_mass.take(link_idx) * (link_idx > -1)

    n = -contact.frame[0]
    rel_vel = (
        xd.vel[0]
        + jp.cross(xd.ang[0], contact.pos - x.pos[0])
        - (xd.vel[1] + jp.cross(xd.ang[1], contact.pos - x.pos[1]))
    )
    v_n = jp.dot(rel_vel, n)
    v_t = rel_vel - n * v_n
    v_t_dir, v_t_norm = math.normalize(v_t)
    dvel = -v_t_dir * jp.minimum(
        contact.friction[0] * jp.abs(dlambda) / sys.opt.timestep, v_t_norm
    )

    angw_1 = jp.cross((contact.pos - x.pos[0]), v_t_dir)
    angw_2 = jp.cross((contact.pos - x.pos[1]), v_t_dir)
    w1 = mass_inv[0] + jp.dot(angw_1, i_inv[0] @ angw_1)
    w2 = mass_inv[1] + jp.dot(angw_2, i_inv[1] @ angw_2)

    p_dyn = dvel / (w1 + w2 + 1e-6)

    # restitution
    rel_vel_prev = (
        xd_prev.vel[0] + jp.cross(xd_prev.ang[0], contact.pos - x.pos[0])
    ) - (xd_prev.vel[1] + jp.cross(xd_prev.ang[1], contact.pos - x.pos[1]))
    v_n_prev = jp.dot(rel_vel_prev, n)
    dv_rest = n * (-v_n - jp.minimum(contact.elasticity * v_n_prev, 0))

    pos_p = contact.pos
    pos_c = contact.pos + contact.frame[0] * contact.dist
    dx = dv_rest
    pos_p, pos_c = pos_p - x.pos[0], pos_c - x.pos[1]

    c = math.safe_norm(dx)
    n = dx / (c + 1e-6)

    # ignoring inertial effects for now
    cr1, cr2 = jp.cross(pos_p, n), jp.cross(pos_c, n)
    w1 = mass_inv[0] + jp.dot(cr1, i_inv[0] @ cr1)
    w2 = mass_inv[1] + jp.dot(cr2, i_inv[1] @ cr2)

    dlambda_rest = c / (w1 + w2 + 1e-6)
    penetrating = contact.dist < 0
    sinking = v_n_prev <= 0.0

    p = Force.create(vel=(dlambda_rest * n * sinking + p_dyn) * penetrating)

    return p, jp.asarray(penetrating, dtype=jp.float32)

  p, is_contact = impulse(contact, dlambda)

  # calculate the impulse to each link center of mass
  p = jax.tree.map(lambda x: jp.concatenate((x, -x)), p)
  pos = jp.tile(contact.pos, (2, 1))
  link_idx = jp.concatenate(contact.link_idx)
  xp_i = Transform.create(pos=pos - x_i.take(link_idx).pos).vmap().do(p)
  xp_i = jax.tree.map(lambda x: segment_sum(x, link_idx, sys.num_links()), xp_i)

  # average the impulse across multiple contacts
  num_contacts = segment_sum(jp.tile(is_contact, 2), link_idx, sys.num_links())
  xp_i = xp_i / (num_contacts.reshape((-1, 1)) + 1e-8)

  # convert impulse to delta-velocity
  xdv_i = Motion(
      vel=jax.vmap(lambda x, y: x * y)(inv_mass, xp_i.vel),
      ang=jax.vmap(lambda x, y: x @ y)(inv_inertia, xp_i.ang),
  )

  return xdv_i



# Copyright 2024 The Brax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Data processing utilities."""
# pylint:disable=g-importing-member
# pylint:disable=g-doc-args
# pylint:disable=g-doc-return-or-yield
from collections import OrderedDict as odict
from typing import Any, Union, Tuple, Dict
from jax import numpy as jnp
import numpy as np


def get_indices(shape: Dict[str, Any]):
  """Check if shape is valid and return indices."""
  s = shape
  indices = s.get('indices', None)
  if indices is None:
    try:
      indices = tuple(range(s['start'], s['end']))
    except KeyError:
      raise NotImplementedError(s)
  assert np.prod(s['shape']) == s['size'], s
  assert len(indices) == s['size'], s
  return indices


def flatten_array(arr: jnp.ndarray, shape: Dict[str, Any]):
  """Flatten array."""
  s = shape
  assert arr.shape[-len(s['shape']
                       ):] == s['shape'], f'{arr.shape} incompat {s["shape"]}'
  arr = arr.reshape(arr.shape[:-len(s['shape'])] + (s['size'],))
  return arr


def check_array_shapes(array_dict: Union[Dict[str, jnp.ndarray], jnp.ndarray],
                       array_shapes: Dict[str, Dict[str, Any]]):
  """Check array_dict compatible with array_shapes."""
  leading_dims = None
  for k, s in array_shapes.items():
    assert k in array_dict, f'{k} not in {list(array_dict.keys())}'
    arr = array_dict[k]
    if leading_dims is None:
      leading_dims = arr.shape[:-len(s['shape'])]
    else:
      assert leading_dims == arr.shape[:-len(
          s['shape'])], f'{k}: {leading_dims} incompat {arr.shape}'
    assert arr.shape[-len(s['shape']
                         ):] == s['shape'], f'{k}: {arr.shape} != {s["shape"]}'


def get_array_shapes(array_dict: Union[Dict[str, jnp.ndarray], jnp.ndarray],
                     batch_shape: Tuple[int] = ()):
  """Get array dict shape.

  get_array_shape() returns shape info in the form:
    dict(key1=dict(shape=(10,), start=40, end=50), ...)
  """
  if isinstance(array_dict, jnp.ndarray):
    return array_dict.shape
  array_shapes = odict()
  i = 0
  for k, v in array_dict.items():
    v_shape = v.shape[len(batch_shape):]
    size = np.prod(v_shape)
    array_shapes[k] = dict(shape=v_shape, size=size, start=i, end=i + size)
    i += size
  return array_shapes


def fill_array(array_dict: Dict[str, jnp.ndarray],
               array: jnp.ndarray,
               array_shapes: Dict[str, Dict[str, Any]],
               assert_once: bool = False) -> jnp.ndarray:
  """Fill an array with array_dict."""
  check_array_shapes(array_dict, array_shapes)
  if assert_once:
    assert array.shape[-1] == sum([s['size'] for s in array_shapes.values()
                                  ]), f'{array.shape} incompat {array_shapes}'
    assigned = jnp.zeros(array.shape[-1])
  for k in array_shapes.keys():
    arr = array_dict[k]
    s = array_shapes[k]
    indices = get_indices(s)
    arr = flatten_array(arr, s)
    array = array.at[..., indices].set(arr)
    if assert_once:
      assigned = assigned.at[..., indices].add(1)
  if assert_once:
    assert jnp.all(jnp.equal(assigned, 1)), assigned
  return array


def concat_array(array_dict: Dict[str, jnp.ndarray],
                 array_shapes: Dict[str, Dict[str, Any]]) -> jnp.ndarray:
  """Concatenate array dictionary to a vector."""
  check_array_shapes(array_dict, array_shapes)
  return jnp.concatenate([
      arr.reshape(arr.shape[:-len(s['shape'])] + (s['size'],))
      for arr, s in zip(array_dict.values(), array_shapes.values())
  ],
                         axis=-1)


def split_array(
    array: jnp.ndarray,
    array_shapes: Dict[str, Dict[str, Any]]) -> Dict[str, jnp.ndarray]:
  """Split array vector to a dictionary."""
  array_leading_dims = array.shape[:-1]
  array_dict = odict()
  for k, s in array_shapes.items():
    indices = get_indices(s)
    array_dict[k] = array[..., indices].reshape(array_leading_dims + s['shape'])
  return array_dict



# Copyright 2024 The Brax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# pylint:disable=g-multiple-import
"""Trains a robot arm to push a ball to a target."""

from brax import base
from brax import math
from brax.envs.base import PipelineEnv, State
from brax.io import mjcf
from etils import epath
import jax
from jax import numpy as jp


class Pusher(PipelineEnv):



  # pyformat: disable
  """
  ### Description

  "Pusher" is a multi-jointed robot arm which is very similar to that of a
  human.

  The goal is to move a target cylinder (called *object*) to a goal position
  using the robot's end effector (called *fingertip*). The robot consists of
  shoulder, elbow, forearm, and wrist joints.

  ### Action Space

  The action space is a `Box(-2, 2, (7,), float32)`. An action `(a, b)`
  represents the torques applied at the hinge joints.

  | Num | Action                                        | Control Min | Control Max | Name (in corresponding config) | Joint | Unit         |
  |-----|-----------------------------------------------|-------------|-------------|--------------------------------|-------|--------------|
  | 0   | Rotation of the panning the shoulder          | -1          | 1           | r_shoulder_pan_joint           | hinge | torque (N m) |
  | 1   | Rotation of the shoulder lifting joint        | -1          | 1           | r_shoulder_lift_joint          | hinge | torque (N m) |
  | 2   | Rotation of the shoulder rolling joint        | -1          | 1           | r_upper_arm_roll_joint         | hinge | torque (N m) |
  | 3   | Rotation of hinge joint that flexed the elbow | -1          | 1           | r_elbow_flex_joint             | hinge | torque (N m) |
  | 4   | Rotation of hinge that rolls the forearm      | -1          | 1           | r_forearm_roll_joint           | hinge | torque (N m) |
  | 5   | Rotation of flexing the wrist                 | -1          | 1           | r_wrist_flex_joint             | hinge | torque (N m) |
  | 6   | Rotation of rolling the wrist                 | -1          | 1           | r_wrist_roll_joint             | hinge | torque (N m) |

  ### Observation Space

  Observations consist of

  - Angle of rotational joints on the pusher
  - Angular velocities of rotational joints on the pusher
  - The coordinates of the fingertip of the pusher
  - The coordinates of the object to be moved
  - The coordinates of the goal position

  The observation is a `ndarray` with shape `(23,)` where the elements
  correspond to the table below. An analogy can be drawn to a human arm in order
  to help understand the state space, with the words flex and roll meaning the
  same as human joints.

  | Num | Observation                                              | Min  | Max | Name (in corresponding config) | Joint    | Unit                     |
  |-----|----------------------------------------------------------|------|-----|--------------------------------|----------|--------------------------|
  | 0   | Rotation of the panning the shoulder                     | -Inf | Inf | r_shoulder_pan_joint           | hinge    | angle (rad)              |
  | 1   | Rotation of the shoulder lifting joint                   | -Inf | Inf | r_shoulder_lift_joint          | hinge    | angle (rad)              |
  | 2   | Rotation of the shoulder rolling joint                   | -Inf | Inf | r_upper_arm_roll_joint         | hinge    | angle (rad)              |
  | 3   | Rotation of hinge joint that flexed the elbow            | -Inf | Inf | r_elbow_flex_joint             | hinge    | angle (rad)              |
  | 4   | Rotation of hinge that rolls the forearm                 | -Inf | Inf | r_forearm_roll_joint           | hinge    | angle (rad)              |
  | 5   | Rotation of flexing the wrist                            | -Inf | Inf | r_wrist_flex_joint             | hinge    | angle (rad)              |
  | 6   | Rotation of rolling the wrist                            | -Inf | Inf | r_wrist_roll_joint             | hinge    | angle (rad)              |
  | 7   | Rotational velocity of the panning the shoulder          | -Inf | Inf | r_shoulder_pan_joint           | hinge    | angular velocity (rad/s) |
  | 8   | Rotational velocity of the shoulder lifting joint        | -Inf | Inf | r_shoulder_lift_joint          | hinge    | angular velocity (rad/s) |
  | 9   | Rotational velocity of the shoulder rolling joint        | -Inf | Inf | r_upper_arm_roll_joint         | hinge    | angular velocity (rad/s) |
  | 10  | Rotational velocity of hinge joint that flexed the elbow | -Inf | Inf | r_elbow_flex_joint             | hinge    | angular velocity (rad/s) |
  | 11  | Rotational velocity of hinge that rolls the forearm      | -Inf | Inf | r_forearm_roll_joint           | hinge    | angular velocity (rad/s) |
  | 12  | Rotational velocity of flexing the wrist                 | -Inf | Inf | r_wrist_flex_joint             | hinge    | angular velocity (rad/s) |
  | 13  | Rotational velocity of rolling the wrist                 | -Inf | Inf | r_wrist_roll_joint             | hinge    | angular velocity (rad/s) |
  | 14  | x-coordinate of the fingertip of the pusher              | -Inf | Inf | tips_arm                       | slide    | position (m)             |
  | 15  | y-coordinate of the fingertip of the pusher              | -Inf | Inf | tips_arm                       | slide    | position (m)             |
  | 16  | z-coordinate of the fingertip of the pusher              | -Inf | Inf | tips_arm                       | slide    | position (m)             |
  | 17  | x-coordinate of the object to be moved                   | -Inf | Inf | object (obj_slidex)            | slide    | position (m)             |
  | 18  | y-coordinate of the object to be moved                   | -Inf | Inf | object (obj_slidey)            | slide    | position (m)             |
  | 19  | z-coordinate of the object to be moved                   | -Inf | Inf | object                         | cylinder | position (m)             |
  | 20  | x-coordinate of the goal position of the object          | -Inf | Inf | goal (goal_slidex)             | slide    | position (m)             |
  | 21  | y-coordinate of the goal position of the object          | -Inf | Inf | goal (goal_slidey)             | slide    | position (m)             |
  | 22  | z-coordinate of the goal position of the object          | -Inf | Inf | goal                           | sphere   | position (m)             |


  ### Rewards

  The reward consists of two parts:

  - *reward_near*: This reward is a measure of how far the *fingertip* of the
  pusher (the unattached end) is from the object, with a more negative value
  assigned for when the pusher's *fingertip* is further away from the target.
  It is calculated as the negative vector norm of (position of the fingertip -
  position of target), or *-norm("fingertip" - "target")*.
  - *reward_dist*: This reward is a measure of how far the object is from the
  target goal position, with a more negative value assigned for object is
  further away from the target. It is calculated as the negative vector norm of
  (position of the object - position of goal), or *-norm("object" - "target")*.
  - *reward_control*: A negative reward for penalising the pusher if it takes
  actions that are too large. It is measured as the negative squared Euclidean
  norm of the action, i.e. as *- sum(action<sup>2</sup>)*.

  Unlike other environments, Pusher does not allow you to specify weights for
  the individual reward terms. However, `info` does contain the keys
  *reward_dist* and *reward_ctrl*. Thus, if you'd like to weight the terms, you
  should create a wrapper that computes the weighted reward from `info`.

  ### Starting State

  All pusher (not including object and goal) states start in (0.0, 0.0, 0.0,
  0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0). A uniform noise in the
  range [-0.005, 0.005] is added to the velocity attributes only. The velocities
  of the object and goal are permanently set to 0. The object's x-position is
  selected uniformly between [-0.3, 0] while the y-position is selected
  uniformly between [-0.2, 0.2], and this process is repeated until the vector
  norm between the object's (x,y) position and origin is not greater than 0.17.
  The goal always have the same position of (0.45, -0.05, -0.323).

  The default *dt = 0.05*.

  ### Episode Termination

  The episode terminates when any of the following happens:

  1. The episode duration reaches a 1000 timesteps.
  """
  # pyformat: enable


  def __init__(self, backend='generalized', **kwargs):
    path = epath.resource_path('brax') / 'envs/assets/pusher.xml'
    sys = mjcf.load(path)

    n_frames = 5

    if backend in ['spring', 'positional']:
      sys = sys.tree_replace({'opt.timestep': 0.001})
      sys = sys.replace(
          actuator=sys.actuator.replace(gear=jp.array([20.0] * sys.act_size()))
      )
      n_frames = 50

    kwargs['n_frames'] = kwargs.get('n_frames', n_frames)

    super().__init__(sys=sys, backend=backend, **kwargs)

    # The tips_arm body gets fused with r_wrist_roll_link, so we use the parent
    # r_wrist_flex_link for tips_arm_idx.
    self._tips_arm_idx = self.sys.link_names.index('r_wrist_flex_link')
    self._object_idx = self.sys.link_names.index('object')
    self._goal_idx = self.sys.link_names.index('goal')

  def reset(self, rng: jax.Array) -> State:
    qpos = self.sys.init_q

    rng, rng1, rng2 = jax.random.split(rng, 3)

    # randomly orient the object
    cylinder_pos = jp.concatenate([
        jax.random.uniform(rng, (1,), minval=-0.3, maxval=-1e-6),
        jax.random.uniform(rng1, (1,), minval=-0.2, maxval=0.2),
    ])
    # constrain minimum distance of object to goal
    goal_pos = jp.array([0.0, 0.0])
    norm = math.safe_norm(cylinder_pos - goal_pos)
    scale = jp.where(norm < 0.17, 0.17 / norm, 1.0)
    cylinder_pos *= scale
    qpos = qpos.at[-4:].set(jp.concatenate([cylinder_pos, goal_pos]))

    qvel = jax.random.uniform(
        rng2, (self.sys.qd_size(),), minval=-0.005, maxval=0.005
    )
    qvel = qvel.at[-4:].set(0.0)

    pipeline_state = self.pipeline_init(qpos, qvel)

    obs = self._get_obs(pipeline_state)
    reward, done, zero = jp.zeros(3)
    metrics = {'reward_dist': zero, 'reward_ctrl': zero, 'reward_near': zero}
    return State(pipeline_state, obs, reward, done, metrics)

  def step(self, state: State, action: jax.Array) -> State:

    # Scale action from [-1,1] to actuator limits
    action_min = self.sys.actuator.ctrl_range[:, 0]
    action_max = self.sys.actuator.ctrl_range[:, 1]
    action = (action + 1) * (action_max - action_min) * 0.5 + action_min

    pipeline_state = self.pipeline_step(state.pipeline_state, action)

    assert state.pipeline_state is not None
    x_i = state.pipeline_state.x.vmap().do(
        base.Transform.create(pos=self.sys.link.inertia.transform.pos)
    )
    vec_1 = x_i.pos[self._object_idx] - x_i.pos[self._tips_arm_idx]
    vec_2 = x_i.pos[self._object_idx] - x_i.pos[self._goal_idx]

    reward_near = -math.safe_norm(vec_1)
    reward_dist = -math.safe_norm(vec_2)
    reward_ctrl = -jp.square(action).sum()
    reward = reward_dist + 0.1 * reward_ctrl + 0.5 * reward_near

    obs = self._get_obs(pipeline_state)
    state.metrics.update(
        reward_near=reward_near,
        reward_dist=reward_dist,
        reward_ctrl=reward_ctrl,
    )
    return state.replace(pipeline_state=pipeline_state, obs=obs, reward=reward)

  def _get_obs(self, pipeline_state: base.State) -> jax.Array:
    """Observes pusher body position and velocities."""
    x_i = pipeline_state.x.vmap().do(
        base.Transform.create(pos=self.sys.link.inertia.transform.pos)
    )
    return jp.concatenate([
        pipeline_state.q[:7],
        pipeline_state.qd[:7],
        x_i.pos[self._tips_arm_idx],
        x_i.pos[self._object_idx],
        x_i.pos[self._goal_idx],
    ])



# Copyright 2024 The Brax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Tests for contacts."""

from absl.testing import absltest
from brax import contact
from brax import kinematics
from brax import test_utils
from brax.io import mjcf
from etils import epath
from jax import numpy as jp
import numpy as np


class SphereTest(absltest.TestCase):

  def test_sphere_plane(self):
    sys = test_utils.load_fixture('ant.xml')
    # all 4 feet are just barely contacting the floor
    q = jp.array([0, 0, 0.556008, 1, 0, 0, 0, 0, 1, 0, 1, 0, 1, 0, 1])
    x, _ = kinematics.forward(sys, q, jp.zeros(sys.qd_size()))
    c = contact.get(sys, x)
    self.assertEqual(c.pos.shape[0], 4)  # 4 contacts

    # front right foot is in +x, +y quadrant and is just resting on the surface
    c = c.take(0)
    np.testing.assert_array_almost_equal(c.pos, jp.array([0.61612, 0.61612, 0]))
    np.testing.assert_array_almost_equal(c.frame[0], jp.array([0, 0, 1]))
    np.testing.assert_array_almost_equal(c.dist, 0)
    np.testing.assert_array_almost_equal(c.friction[0], 1)
    np.testing.assert_array_almost_equal(c.elasticity, 0)
    self.assertEqual(c.link_idx, (-1, 2))

  _SPHERE_SPHERE = """
    <mujoco model="two_spheres">
      <custom>
        <numeric data="0.2" name="elasticity"/>
        <tuple name="elasticity">
          <element objtype="geom" objname="sphere1" prm="0.1"/>
        </tuple>
      </custom>
      <worldbody>
        <body name="body1" pos="0 0 0">
          <joint axis="1 0 0" name="free1" pos="0 0 0" type="free"/>
          <geom name="sphere1" pos="0 0 0" size="0.2" type="sphere"/>
        </body>
        <body name="body2" pos="0 0 0">
          <joint axis="1 0 0" name="free2" pos="0 0 0" type="free"/>
          <geom name="sphere2" pos="0 0.3 0" size="0.11" type="sphere"/>
        </body>
      </worldbody>
    </mujoco>
  """

  def test_sphere_sphere(self):
    sys = mjcf.loads(self._SPHERE_SPHERE)
    x, _ = kinematics.forward(sys, sys.init_q, jp.zeros(sys.qd_size()))
    c = contact.get(sys, x).take(0)

    np.testing.assert_array_almost_equal(c.dist, -0.01)
    np.testing.assert_array_almost_equal(c.pos, jp.array([0.0, 0.195, 0.0]))
    np.testing.assert_array_almost_equal(c.frame[0], jp.array([0, 1.0, 0.0]))
    np.testing.assert_array_almost_equal(c.elasticity, jp.array([0.15]))


class ConvexTest(absltest.TestCase):

  _CONVEX_CONVEX = """
    <mujoco model="convex_convex">
      <asset>
        <mesh name="tetrahedron" file="meshes/tetrahedron.stl" scale="0.1 0.1 0.1" />
        <mesh name="dodecahedron" file="meshes/dodecahedron.stl" scale="0.01 0.01 0.01" />
      </asset>
      <worldbody>
        <body name="obj1" pos="0.0 2.0 0.096">
          <joint axis="1 0 0" name="free1" pos="0 0 0" type="free"/>
          <geom name="obj1" pos="0 0 0" size="0.2 0.2 0.2" type="mesh" mesh="tetrahedron"/>
        </body>
        <body name="obj2" pos="0.0 2.0 0.289" euler="0.1 -0.1 45">
          <joint axis="1 0 0" name="free2" pos="0 0 0" type="free"/>
          <geom name="obj2" pos="0 0 0" size="0.1 0.1 0.1" type="mesh" mesh="dodecahedron"/>
        </body>
      </worldbody>
    </mujoco>
  """

  def test_convex_convex(self):
    """Tests generic convex-convex collision."""
    sys = mjcf.loads(
        self._CONVEX_CONVEX,
        asset_path=epath.resource_path('brax') / 'test_data/',
    )
    x, _ = kinematics.forward(sys, sys.init_q, jp.zeros(sys.qd_size()))
    c = contact.get(sys, x)
    # Only one contact point for an edge contact.
    self.assertLess(c.dist[0], 0)
    self.assertTrue((c.dist[1:] > 0).all())
    np.testing.assert_array_almost_equal(c.frame[0, 0], jp.array([0, 0, 1]))


class GeomPairTest(absltest.TestCase):
  """Tests for colliding geom pairs."""

  def test_no_world_self_collision(self):
    sys = test_utils.load_fixture('world_self_collision.xml')
    x, _ = kinematics.forward(sys, sys.init_q, jp.zeros(sys.qd_size()))
    c = contact.get(sys, x)
    self.assertEqual(c.pos.shape[0], 13)

  def test_collision_with_world_geom(self):
    sys = test_utils.load_fixture('world_fromto.xml')
    x, _ = kinematics.forward(sys, sys.init_q, jp.zeros(sys.qd_size()))
    c = contact.get(sys, x)
    self.assertEqual(c.pos.shape[0], 1)


if __name__ == '__main__':
  absltest.main()



# Copyright 2024 The Brax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Command line tool for converting Mujoco models to Brax."""

from typing import Sequence

from absl import app
from absl import flags
from absl import logging
from brax.v1.io import file
from brax.v1.tools import mujoco

from google.protobuf import text_format

FLAGS = flags.FLAGS

flags.DEFINE_string('xml_model_path', None,
                    'Path of the Mujoco XML model to import.')
flags.DEFINE_string('config_path', None, 'Path of the output config.')
flags.DEFINE_bool('add_collision_pairs', True,
                  'Adds the collision pairs to the config.')
# System parameters. See brax/physics/config.proto for more information.
flags.DEFINE_float('angular_damping', -0.05,
                   'Angular velocity damping applied to each body.')
flags.DEFINE_float(
    'baumgarte_erp', 0.1,
    'How aggressively interpenetrating bodies should push away each another.')
flags.DEFINE_float('dt', 0.02, 'Time to simulate each step, in seconds.')
flags.DEFINE_float('friction', 0.6,
                   'How much surfaces in contact resist translation.')
flags.DEFINE_integer('substeps', 4,
                     'Substeps to perform to maintain numerical stability.')
flags.DEFINE_bool('ignore_unsupported_joints', False,
                  'Ignores unsupported joints.')
flags.DEFINE_bool('add_joint_to_nearest_body', False,
                  'Adds a joint to the nearest (child)body.')


def main(argv: Sequence[str]) -> None:
  if len(argv) > 1:
    raise app.UsageError('Too many command-line arguments.')

  # Read the Mujoco model.
  filename = FLAGS.xml_model_path
  with file.File(filename) as f:
    logging.info('Loading mujoco model from %s', filename)
    xml_string = f.read()

  # Convert the model.
  m = mujoco.MujocoConverter(
      xml_string,
      add_collision_pairs=FLAGS.add_collision_pairs,
      ignore_unsupported_joints=FLAGS.ignore_unsupported_joints,
      add_joint_to_nearest_body=FLAGS.add_joint_to_nearest_body)
  config = m.config

  # Add the default options.
  config.angular_damping = FLAGS.angular_damping
  config.baumgarte_erp = FLAGS.baumgarte_erp
  config.dt = FLAGS.dt
  config.friction = FLAGS.friction
  config.substeps = FLAGS.substeps

  # Save the config.
  if FLAGS.config_path:
    text_proto = text_format.MessageToString(config)
    with file.File(FLAGS.config_path, mode='w+') as f:
      f.write(text_proto)


if __name__ == '__main__':
  app.run(main)



# Copyright 2024 The Brax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Loading/saving of inference functions."""

import pickle
from typing import Any

from brax.v1.io.file import File


def load_params(path: str) -> Any:
  with File(path, 'rb') as fin:
    buf = fin.read()
  return pickle.loads(buf)


def save_params(path: str, params: Any):
  """Saves parameters in Flax format."""
  with File(path, 'wb') as fout:
    fout.write(pickle.dumps(params))



# Copyright 2024 The Brax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Helper functions for physics calculations in maximal coordinates."""
# pylint:disable=g-multiple-import
from typing import Tuple

from brax import math
from brax.base import Motion, System, Transform
import jax
from jax import numpy as jp


def from_world(
    sys: System, x: Transform, xd: Motion
) -> Tuple[Transform, Motion]:
  """Converts link transform and motion from world frame to com frame."""
  x_i = x.vmap().do(Transform.create(pos=sys.link.inertia.transform.pos))
  xd_i = Transform.create(pos=x_i.pos - x.pos).vmap().do(xd)
  return x_i, xd_i


def to_world(
    sys: System, x_i: Transform, xd_i: Motion
) -> Tuple[Transform, Motion]:
  """Converts link transform and motion from com frame to world frame."""
  x = x_i.vmap().do(Transform.create(pos=-sys.link.inertia.transform.pos))
  xd = Transform.create(pos=x.pos - x_i.pos).vmap().do(xd_i)
  return x, xd


def inv_inertia(sys, x) -> jax.Array:
  """Gets the inverse inertia at the center of mass in world frame."""

  @jax.vmap
  def inv_i(link_inertia, x_rot):
    ri = math.quat_mul(x_rot, link_inertia.transform.rot)
    i_diag = jp.diagonal(link_inertia.i) ** (1 - sys.spring_inertia_scale)
    i_inv_mx = jp.diag(1 / i_diag)
    i_rot_row = jax.vmap(math.rotate, in_axes=[0, None])(i_inv_mx, ri)
    i_rot_col = jax.vmap(math.rotate, in_axes=[0, None])(i_rot_row.T, ri)
    return i_rot_col

  return inv_i(sys.link.inertia, x.rot)



# Copyright 2024 The Brax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# pylint:disable=g-multiple-import, g-importing-member
"""Functions for forces/torques through fluids."""

from typing import Union
from brax.base import Force, Motion, System, Transform
import jax
import jax.numpy as jp


def _box_viscosity(box: jax.Array, xd_i: Motion, viscosity: jax.Array) -> Force:
  """Gets force due to motion through a viscous fluid."""
  diam = jp.mean(box, axis=-1)
  ang_scale = -jp.pi * diam**3 * viscosity
  vel_scale = -3.0 * jp.pi * diam * viscosity
  frc = Force(
      ang=ang_scale[:, None] * xd_i.ang, vel=vel_scale[:, None] * xd_i.vel
  )
  return frc


def _box_density(box: jax.Array, xd_i: Motion, density: jax.Array) -> Force:
  """Gets force due to motion through dense fluid."""

  @jax.vmap
  def apply(b: jax.Array, xd: Motion) -> Force:
    box_mult_vel = jp.array([b[1] * b[2], b[0] * b[2], b[0] * b[1]])
    vel = -0.5 * density * box_mult_vel * jp.abs(xd.vel) * xd.vel
    box_mult_ang = jp.array([
        b[0] * (b[1] ** 4 + b[2] ** 4),
        b[1] * (b[0] ** 4 + b[2] ** 4),
        b[2] * (b[0] ** 4 + b[1] ** 4),
    ])
    ang = -1.0 * density * box_mult_ang * jp.abs(xd.ang) * xd.ang / 64.0
    return Force(vel=vel, ang=ang)

  return apply(box, xd_i)


def force(
    sys: System,
    x: Transform,
    xd: Motion,
    mass: jax.Array,
    inertia: jax.Array,
    root_com: Union[jax.Array, None] = None,
) -> Force:
  """Returns force due to motion through a fluid."""
  # get the velocity at the com position/orientation
  x_i = x.vmap().do(sys.link.inertia.transform)
  # TODO: remove root_com when xd is fixed for stacked joints
  offset = x_i.pos - x.pos if root_com is None else x_i.pos - root_com
  xd_i = x_i.replace(pos=offset).vmap().do(xd)

  # TODO: add ellipsoid fluid model from mujoco
  # TODO: consider adding wind from mj.opt.wind
  diag_inertia = jax.vmap(jp.diag)(inertia)
  diag_inertia_v = jp.repeat(diag_inertia, 3, axis=-2).reshape((-1, 3, 3))
  diag_inertia_v *= jp.ones((3, 3)) - 2 * jp.eye(3)
  box = 6.0 * jp.clip(jp.sum(diag_inertia_v, axis=-1), a_min=1e-12)
  box = jp.sqrt(box / mass[:, None])

  frc = _box_viscosity(box, xd_i, sys.viscosity)
  frc += _box_density(box, xd_i, sys.density)

  # rotate back to the world orientation
  frc = Transform.create(rot=x_i.rot).vmap().do(frc)

  return frc



# Copyright 2024 The Brax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# pylint:disable=g-multiple-import
"""An inverted pendulum environment."""

from brax import base
from brax.envs.base import PipelineEnv, State
from brax.io import mjcf
from etils import epath
import jax
from jax import numpy as jp


class InvertedDoublePendulum(PipelineEnv):



  # pyformat: disable
  """### Description

  This environment originates from control theory and builds on the cartpole
  environment based on the work done by Barto, Sutton, and Anderson in
  ["Neuronlike adaptive elements that can solve difficult learning control
  problems"](https://ieeexplore.ieee.org/document/6313077).

  This environment involves a cart that can moved linearly, with a pole fixed on
  it and a second pole fixed on the other end of the first one (leaving the
  second pole as the only one with one free end). The cart can be pushed left or
  right, and the goal is to balance the second pole on top of the first pole,
  which is in turn on top of the cart, by applying continuous forces on the
  cart.

  ### Action Space

  The agent take a 1-element vector for actions.

  The action space is a continuous `(action)` in `[-1, 1]`, where `action`
  represents the numerical force applied to the cart (with magnitude
  representing the amount of force and sign representing the direction)

  | Num | Action                    | Control Min | Control Max | Name (in
  corresponding config) | Joint | Unit      |
  |-----|---------------------------|-------------|-------------|--------------------------------|-------|-----------|
  | 0   | Force applied on the cart | -1          | 1           | slider
  | slide | Force (N) |

  ### Observation Space

  The state space consists of positional values of different body parts of the
  pendulum system, followed by the velocities of those individual parts (their
  derivatives) with all the positions ordered before all the velocities.

  The observation is a `ndarray` with shape `(11,)` where the elements
  correspond to the following:

  | Num | Observation                                                       |
  Min  | Max | Name (in corresponding config) | Joint | Unit
  |
  |-----|-------------------------------------------------------------------|------|-----|--------------------------------|-------|--------------------------|
  | 0   | position of the cart along the linear surface                     |
  -Inf | Inf | thruster                       | slide | position (m)
  |
  | 1   | sine of the angle between the cart and the first pole             |
  -Inf | Inf | sin(hinge)                     | hinge | unitless
  |
  | 2   | sine of the angle between the two poles                           |
  -Inf | Inf | sin(hinge2)                    | hinge | unitless
  |
  | 3   | cosine of the angle between the cart and the first pole           |
  -Inf | Inf | cos(hinge)                     | hinge | unitless
  |
  | 4   | cosine of the angle between the two poles                         |
  -Inf | Inf | cos(hinge2)                    | hinge | unitless
  |
  | 5   | velocity of the cart                                              |
  -Inf | Inf | thruster                       | slide | velocity (m/s)
  |
  | 6   | angular velocity of the angle between the cart and the first pole |
  -Inf | Inf | hinge                          | hinge | angular velocity (rad/s)
  |
  | 7   | angular velocity of the angle between the two poles               |
  -Inf | Inf | hinge2                         | hinge | angular velocity (rad/s)
  |

  ### Rewards

  The goal is to make the inverted pendulum stand upright (within a certain
  angle limit) as long as possible - as such a reward of +1 is awarded for each
  timestep that the pole is upright.

  ### Starting State

  All observations start in state (0.0, 0.0, 0.0, 0.0) with a uniform noise in
  the range of [-0.01, 0.01] added to the values for stochasticity.

  ### Episode Termination

  The episode terminates when any of the following happens:

  1. The episode duration reaches 1000 timesteps.
  2. The absolute value of the vertical angle between the pole and the cart is
  greater than 0.2 radians.
  """
  # pyformat: enable


  def __init__(self, backend='generalized', **kwargs):
    path = (
        epath.resource_path('brax')
        / 'envs/assets/inverted_double_pendulum.xml'
    )
    sys = mjcf.load(path)

    n_frames = 2

    if backend in ['spring', 'positional']:
      sys = sys.tree_replace({'opt.timestep': 0.005})
      n_frames = 4

    kwargs['n_frames'] = kwargs.get('n_frames', n_frames)

    super().__init__(sys=sys, backend=backend, **kwargs)

  def reset(self, rng: jax.Array) -> State:
    """Resets the environment to an initial state."""
    rng, rng1, rng2 = jax.random.split(rng, 3)

    q = self.sys.init_q + jax.random.uniform(
        rng1, (self.sys.q_size(),), minval=-0.01, maxval=0.01
    )
    qd = jax.random.normal(rng2, (self.sys.qd_size(),)) * 0.01
    pipeline_state = self.pipeline_init(q, qd)

    obs = self._get_obs(pipeline_state)
    reward, done = jp.zeros(2)
    metrics = {}

    return State(pipeline_state, obs, reward, done, metrics)

  def step(self, state: State, action: jax.Array) -> State:
    """Run one timestep of the environment's dynamics."""
    pipeline_state = self.pipeline_step(state.pipeline_state, action)

    tip = base.Transform.create(pos=jp.array([0.0, 0.0, 0.6])).do(
        pipeline_state.x.take(2)
    )
    x, _, y = tip.pos
    dist_penalty = 0.01 * x**2 + (y - 2) ** 2
    v1, v2 = pipeline_state.qd[1:]
    vel_penalty = 1e-3 * v1**2 + 5e-3 * v2**2
    alive_bonus = 10

    obs = self._get_obs(pipeline_state)
    reward = alive_bonus - dist_penalty - vel_penalty
    done = jp.where(y <= 1, jp.float32(1), jp.float32(0))

    return state.replace(
        pipeline_state=pipeline_state, obs=obs, reward=reward, done=done
    )

  @property
  def action_size(self):
    return 1

  def _get_obs(self, pipeline_sate: base.State) -> jax.Array:
    """Observe cartpole body position and velocities."""
    return jp.concatenate(
        [
            pipeline_sate.q[:1],  # cart x pos
            jp.sin(pipeline_sate.q[1:]),
            jp.cos(pipeline_sate.q[1:]),
            jp.clip(pipeline_sate.qd, -10, 10),
            # qfrc_constraint is not added
        ]
    )



# Copyright 2024 The Brax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# pylint:disable=redefined-builtin
"""Common math functions used by multiple brax modules."""
from typing import Tuple

from brax.v1 import jumpy as jp

Vector3 = jp.ndarray
Quaternion = jp.ndarray


def rotate(vec: Vector3, quat: Quaternion) -> Vector3:
  """Rotates a vector vec by a unit quaternion quat.

  Args:
    vec: (3,) a vector
    quat: (4,) a quaternion

  Returns:
    ndarray(3) containing vec rotated by quat.
  """
  if len(vec.shape) != 1:
    raise AssertionError('vec must have no batch dimensions.')
  s, u = quat[0], quat[1:]
  r = 2 * (jp.dot(u, vec) * u) + (s * s - jp.dot(u, u)) * vec
  r = r + 2 * s * jp.cross(u, vec)
  return r


def inv_rotate(vec: Vector3, quat: Quaternion) -> Vector3:
  """Rotates a vector by the inverse of a unit quaternion.

  Args:
    vec: (3,) a vector
    quat: (4,) a quaternion

  Returns:
    A vector rotated by quat^{-1}
  """
  return rotate(vec, quat_inv(quat))


def ang_to_quat(ang: Vector3) -> Quaternion:
  """Converts angular velocity to a quaternion.

  Args:
    ang: (3,) angular velocity

  Returns:
    A rotation quaternion.
  """
  return jp.array([0, ang[0], ang[1], ang[2]])


def euler_to_quat(v: Vector3) -> Quaternion:
  """Converts euler rotations in degrees to quaternion."""
  # this follows the Tait-Bryan intrinsic rotation formalism: x-y'-z''
  c1, c2, c3 = jp.cos(v * jp.pi / 360)
  s1, s2, s3 = jp.sin(v * jp.pi / 360)
  w = c1 * c2 * c3 - s1 * s2 * s3
  x = s1 * c2 * c3 + c1 * s2 * s3
  y = c1 * s2 * c3 - s1 * c2 * s3
  z = c1 * c2 * s3 + s1 * s2 * c3
  return jp.array([w, x, y, z])


def quat_to_euler(q: Quaternion) -> Vector3:
  """Converts quaternions to euler rotations in radians."""
  # this follows the Tait-Bryan intrinsic rotation formalism: x-y'-z''

  z = jp.arctan2(-2 * q[1] * q[2] + 2 * q[0] * q[3],
                 q[1] * q[1] + q[0] * q[0] - q[3] * q[3] - q[2] * q[2])
  # TODO: Investigate why quaternions go so big we need to clip.
  y = jp.safe_arcsin(jp.clip(2 * q[1] * q[3] + 2 * q[0] * q[2], -1., 1.))  # pytype: disable=wrong-arg-types  # jax-ndarray
  x = jp.arctan2(-2 * q[2] * q[3] + 2 * q[0] * q[1],
                 q[3] * q[3] - q[2] * q[2] - q[1] * q[1] + q[0] * q[0])

  return jp.array([x, y, z])


def quat_to_axis_angle(q: Quaternion) -> Tuple[Vector3, jp.ndarray]:
  """Returns the axis-angle representation of a quaternion.

  Args:
    q: (4,) a quaternion

  Returns:
    The angle of axis-angle of this quaternion, in the range [-pi, pi].
  """
  # TODO: replace with more accurate safe function
  # avoid the singularity at 0:
  epsilon = 1e-10
  # safety 1e-10 jitter added because both sqrt and arctan2 have bad gradients
  denom = jp.safe_norm(q[1:])
  angle = 2. * jp.arctan2(
      jp.sqrt(epsilon + q[1] * q[1] + q[2] * q[2] + q[3] * q[3]), q[0])
  angle += jp.where(angle > jp.pi, x=-2 * jp.pi, y=0)  # pytype: disable=wrong-arg-types  # jax-ndarray
  angle += jp.where(angle < -jp.pi, x=2 * jp.pi, y=0)  # pytype: disable=wrong-arg-types  # jax-ndarray
  scale = jp.where(denom == 0., 0., 1. / denom)  # pytype: disable=wrong-arg-types  # jax-ndarray
  return q[1:] * scale, angle


def signed_angle(axis: Vector3, ref_p: Vector3, ref_c: Vector3) -> jp.ndarray:
  """Calculates the signed angle between two vectors along an axis.

  Args:
    axis: (3,) common axis around which to calculate change in angle
    ref_p: (3,) vector pointing at 0-degrees offset in the parent's frame
    ref_c: (3,) vector pointing at 0-degrees offset in the child's frame

  Returns:
    The signed angle between two parts.
  """
  return jp.arctan2(jp.dot(jp.cross(ref_p, ref_c), axis), jp.dot(ref_p, ref_c))


def quat_mul(u: Quaternion, v: Quaternion) -> Quaternion:
  """Multiplies two quaternions.

  Args:
    u: (4,) quaternion (w,x,y,z)
    v: (4,) quaternion (w,x,y,z)

  Returns:
    A quaternion u * v.
  """
  return jp.array([
      u[0] * v[0] - u[1] * v[1] - u[2] * v[2] - u[3] * v[3],
      u[0] * v[1] + u[1] * v[0] + u[2] * v[3] - u[3] * v[2],
      u[0] * v[2] - u[1] * v[3] + u[2] * v[0] + u[3] * v[1],
      u[0] * v[3] + u[1] * v[2] - u[2] * v[1] + u[3] * v[0],
  ])


def vec_quat_mul(u: Vector3, v: Quaternion) -> Quaternion:
  """Multiplies a vector and a quaternion.

  This is a convenience method for multiplying two quaternions when
  one of the quaternions has a 0-value w-part, i.e.:
  quat_mul([0.,a,b,c], [d,e,f,g])

  It is slightly more efficient than constructing a 0-w-part quaternion
  from the vector.

  Args:
    u: (3,) vector representation of the quaternion (0.,x,y,z)
    v: (4,) quaternion (w,x,y,z)

  Returns:
    A quaternion u * v.
  """
  return jp.array([
      -u[0] * v[1] - u[1] * v[2] - u[2] * v[3],
      u[0] * v[0] + u[1] * v[3] - u[2] * v[2],
      -u[0] * v[3] + u[1] * v[0] + u[2] * v[1],
      u[0] * v[2] - u[1] * v[1] + u[2] * v[0],
  ])


def quat_rot_axis(axis: Vector3, angle: jp.ndarray) -> Quaternion:
  """Provides a quaternion that describes rotating around axis v by angle.

  Args:
    axis: (3,) axis (x,y,z)
    angle: () float angle to rotate by

  Returns:
    A quaternion that rotates around v by angle
  """
  qx = axis[0] * jp.sin(angle / 2)
  qy = axis[1] * jp.sin(angle / 2)
  qz = axis[2] * jp.sin(angle / 2)
  qw = jp.cos(angle / 2)
  return jp.array([qw, qx, qy, qz])


def quat_inv(q: Quaternion) -> Quaternion:
  """Calculates the inverse of quaternion q.

  Args:
    q: (4,) quaternion [w, x, y, z]

  Returns:
    The inverse of q, where qmult(q, inv_quat(q)) = [1, 0, 0, 0].
  """
  return q * jp.array([1, -1, -1, -1])


def relative_quat(q1: Quaternion, q2: Quaternion) -> Quaternion:
  """Returns the relative quaternion from q1 to q2."""
  return quat_mul(q2, quat_inv(q1))


def normalize(v: Vector3, epsilon=1e-6) -> Vector3:
  """Normalizes a vector."""
  return v / (epsilon + jp.safe_norm(v))



# Copyright 2024 The Brax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Trains a hopper to run in the +x direction."""

from typing import Tuple

import brax.v1 as brax
from brax.v1 import jumpy as jp
from brax.v1 import math
from brax.v1.envs import env


class Hopper(env.Env):



  """
  ### Description

  This environment is based on the work done by Erez, Tassa, and Todorov in
  ["Infinite Horizon Model Predictive Control for Nonlinear Periodic Tasks"](http://www.roboticsproceedings.org/rss07/p10.pdf). The environment aims to
  increase the number of independent state and control variables as compared to
  the classic control environments.

  The hopper is a two-dimensional one-legged figure that consist of four main
  body parts - the torso at the top, the thigh in the middle, the leg in the
  bottom, and a single foot on which the entire body rests.

  The goal is to make hops that move in the forward (right) direction by
  applying torques on the three hinges connecting the four body parts.

  ### Action Space

  The agent take a 3-element vector for actions. The action space is a
  continuous `(action, action, action)` all in `[-1, 1]`, where `action`
  represents the numerical torques applied between *links*

  | Num | Action                             | Control Min | Control Max | Name (in corresponding config) | Joint | Unit         |
  |-----|------------------------------------|-------------|-------------|--------------------------------|-------|--------------|
  | 0   | Torque applied on the thigh rotor  | -1          | 1           | thigh_joint                    | hinge | torque (N m) |
  | 1   | Torque applied on the leg rotor    | -1          | 1           | leg_joint                      | hinge | torque (N m) |
  | 3   | Torque applied on the foot rotor   | -1          | 1           | foot_joint                     | hinge | torque (N m) |

  ### Observation Space

  The state space consists of positional values of different body parts of the
  hopper, followed by the velocities of those individual parts (their
  derivatives) with all the positions ordered before all the velocities.

  The observation is a `ndarray` with shape `(11,)` where the elements
  correspond to the following:

  | Num | Observation                                      | Min  | Max | Name (in corresponding config) | Joint | Unit                     |
  |-----|--------------------------------------------------|------|-----|--------------------------------|-------|--------------------------|
  | 0   | z-coordinate of the top (height of hopper)       | -Inf | Inf | rootz                          | slide | position (m)             |
  | 1   | angle of the top                                 | -Inf | Inf | rooty                          | hinge | angle (rad)              |
  | 2   | angle of the thigh joint                         | -Inf | Inf | thigh_joint                    | hinge | angle (rad)              |
  | 3   | angle of the leg joint                           | -Inf | Inf | leg_joint                      | hinge | angle (rad)              |
  | 4   | angle of the foot joint                          | -Inf | Inf | foot_joint                     | hinge | angle (rad)              |
  | 5   | velocity of the x-coordinate of the top          | -Inf | Inf | rootx                          | slide | velocity (m/s)           |
  | 6   | velocity of the z-coordinate (height) of the top | -Inf | Inf | rootz                          | slide | velocity (m/s)           |
  | 7   | angular velocity of the angle of the top         | -Inf | Inf | rooty                          | hinge | angular velocity (rad/s) |
  | 8   | angular velocity of the thigh hinge              | -Inf | Inf | thigh_joint                    | hinge | angular velocity (rad/s) |
  | 9   | angular velocity of the leg hinge                | -Inf | Inf | leg_joint                      | hinge | angular velocity (rad/s) |
  | 10  | angular velocity of the foot hinge               | -Inf | Inf | foot_joint                     | hinge | angular velocity (rad/s) |

  ### Rewards

  The reward consists of three parts:

  - *reward_healthy*: Every timestep that the hopper is alive, it gets a reward
    of 1,
  - *reward_forward*: A reward of hopping forward which is measured as
    *(x-coordinate before action - x-coordinate after action)/dt*. *dt* is the
    time between actions - the default *dt = 0.008*. This reward would be
    positive if the hopper hops forward (right) desired.
  - *reward_ctrl*: A negative reward for penalising the hopper if it takes
    actions that are too large. It is measured as *-coefficient **x**
    sum(action<sup>2</sup>)* where *coefficient* is a parameter set for the
    control and has a default value of 0.001

  ### Starting State

  All observations start in state (0.0, 1.25, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0,
  0.0, 0.0) with a uniform noise in the range of [-0.005, 0.005] added to the
  values for stochasticity.

  ### Episode Termination

  The episode terminates when any of the following happens:

  1. The episode duration reaches a 1000 timesteps
  2. The height of the hopper becomes less than 0.7 metres (hopper has hopped
     too low).
  3. The absolute value of the angle (index 2) is less than 0.2 radians (hopper
     has fallen down).

  ### Arguments

  No additional arguments are currently supported (in v2 and lower), but
  modifications can be made to the XML file in the assets folder
  (or by changing the path to a modified XML file in another folder).

  ```
  env = gym.make('Hopper-v2')
  ```

  v3, v4, and v5 take gym.make kwargs such as ctrl_cost_weight,
  reset_noise_scale etc.

  ```
  env = gym.make('Hopper-v5', ctrl_cost_weight=0.1, ....)
  ```

  ### Version History

  * v5: ported to Brax.
  * v4: all mujoco environments now use the mujoco bindings in mujoco>=2.1.3
  * v3: support for gym.make kwargs such as xml_file, ctrl_cost_weight,
        reset_noise_scale etc. rgb rendering comes from tracking camera (so
        agent does not run away from screen)
  * v2: All continuous control environments now use mujoco_py >= 1.50
  * v1: max_time_steps raised to 1000 for robot based tasks. Added
        reward_threshold to environments.
  * v0: Initial versions release (1.0.0)
  """


  def __init__(self,
               forward_reward_weight: float = 1.0,
               ctrl_cost_weight: float = 1e-3,
               healthy_reward: float = 1.0,
               terminate_when_unhealthy: bool = True,
               healthy_z_range: Tuple[float, float] = (0.7, float('inf')),
               healthy_angle_range=(-0.2, 0.2),
               reset_noise_scale=5e-3,
               exclude_current_positions_from_observation=True,
               legacy_spring=False,
               **kwargs):
    """Creates a Hopper environment.

    Args:
      forward_reward_weight: Weight for the forward reward, i.e. velocity in
        x-direction.
      ctrl_cost_weight: Weight for the control cost.
      healthy_reward: Reward for staying healthy, i.e. respecting the posture
        constraints.
      terminate_when_unhealthy: Done bit will be set when unhealthy if true.
      healthy_z_range: Range of the z-position for being healthy.
      healthy_angle_range: Range of joint angles for being healthy.
      reset_noise_scale: Scale of noise to add to reset states.
      exclude_current_positions_from_observation: x-position will be hidden
        from the observations if true.
      legacy_spring: if True, reverts to legacy spring dynamics instead of pbd.
      **kwargs: Arguments that are passed to the base class.
    """
    config = _SYSTEM_CONFIG_SPRING if legacy_spring else _SYSTEM_CONFIG
    super().__init__(config=config, **kwargs)

    self._forward_reward_weight = forward_reward_weight
    self._ctrl_cost_weight = ctrl_cost_weight
    self._healthy_reward = healthy_reward
    self._terminate_when_unhealthy = terminate_when_unhealthy
    self._healthy_z_range = healthy_z_range
    self._healthy_angle_range = healthy_angle_range
    self._reset_noise_scale = reset_noise_scale
    self._exclude_current_positions_from_observation = (
        exclude_current_positions_from_observation
    )

  def reset(self, rng: jp.ndarray) -> env.State:
    """Resets the environment to an initial state."""
    rng, rng1, rng2 = jp.random_split(rng, 3)

    qpos = self.sys.default_angle() + self._noise(rng1)
    qvel = self._noise(rng2)

    qp = self.sys.default_qp(joint_angle=qpos, joint_velocity=qvel)
    obs = self._get_obs(qp)
    reward, done, zero = jp.zeros(3)
    metrics = {
        'reward_forward': zero,
        'reward_ctrl': zero,
        'reward_healthy': zero,
        'x_position': zero,
        'x_velocity': zero,
    }
    return env.State(qp, obs, reward, done, metrics)

  def step(self, state: env.State, action: jp.ndarray) -> env.State:
    """Run one timestep of the environment's dynamics."""
    qp, _ = self.sys.step(state.qp, action)

    x_velocity = (qp.pos[0, 0] - state.qp.pos[0, 0]) / self.sys.config.dt
    forward_reward = self._forward_reward_weight * x_velocity

    min_z, max_z = self._healthy_z_range
    min_angle, max_angle = self._healthy_angle_range
    ang_y = math.quat_to_euler(qp.rot[0])[1]
    is_healthy = jp.where(qp.pos[0, 2] < min_z, x=0.0, y=1.0)  # pytype: disable=wrong-arg-types  # jax-ndarray
    is_healthy = jp.where(qp.pos[0, 2] > max_z, x=0.0, y=is_healthy)  # pytype: disable=wrong-arg-types  # jax-ndarray
    is_healthy = jp.where(ang_y > max_angle, x=0.0, y=is_healthy)  # pytype: disable=wrong-arg-types  # jax-ndarray
    is_healthy = jp.where(ang_y < min_angle, x=0.0, y=is_healthy)  # pytype: disable=wrong-arg-types  # jax-ndarray
    if self._terminate_when_unhealthy:
      healthy_reward = self._healthy_reward
    else:
      healthy_reward = self._healthy_reward * is_healthy

    ctrl_cost = self._ctrl_cost_weight * jp.sum(jp.square(action))

    obs = self._get_obs(qp)
    reward = forward_reward + healthy_reward - ctrl_cost
    done = 1.0 - is_healthy if self._terminate_when_unhealthy else 0.0
    state.metrics.update(
        reward_forward=forward_reward,
        reward_ctrl=-ctrl_cost,
        reward_healthy=healthy_reward,
        x_position=qp.pos[0, 0],
        x_velocity=x_velocity)

    return state.replace(qp=qp, obs=obs, reward=reward, done=done)

  def _get_obs(self, qp: brax.QP) -> jp.ndarray:
    """Returns the environment observations."""
    joint_angle, joint_vel = self.sys.joints[0].angle_vel(qp)
    ang_y = math.quat_to_euler(qp.rot[0])[1:2]

    # qpos: position and orientation of the torso and the joint angles
    if self._exclude_current_positions_from_observation:
      qpos = [qp.pos[0, 2:], ang_y, joint_angle]
    else:
      qpos = [qp.pos[0, (0, 2)], ang_y, joint_angle]

    # qvel: velocity of the torso and the joint angle velocities
    qvel = [qp.vel[0, (0, 2)], qp.ang[0, 1:2], joint_vel]

    return jp.concatenate(qpos + qvel)

  def _noise(self, rng):
    low, hi = -self._reset_noise_scale, self._reset_noise_scale
    return jp.random_uniform(rng, (self.sys.num_joint_dof,), low, hi)


_SYSTEM_CONFIG = """
  bodies {
    name: "torso"
    colliders {
      capsule {
        radius: 0.05
        length: 0.5
      }
    }
    inertia { x: 1.0 y: 1.0 z: 1.0 }
    mass: 3.6651914
  }
  bodies {
    name: "thigh"
    colliders {
      position { z: -0.225 }
      rotation {}
      capsule {
        radius: 0.05
        length: 0.55
      }
    }
    inertia { x: 1.0 y: 1.0 z: 1.0 }
    mass: 4.0578904
  }
  bodies {
    name: "leg"
    colliders {
      position {}
      rotation {}
      capsule {
        radius: 0.04
        length: 0.58
      }
    }
    inertia { x: 1.0 y: 1.0 z: 1.0 }
    mass: 2.7813568
  }
  bodies {
    name: "foot"
    colliders {
      position {
        x: -0.065
        y: -0.13
        z: -0.03
      }
      rotation { y: 90.0 }
      capsule {
        radius: 0.06
        length: 0.51
      }
    }
    inertia { x: 1.0 y: 1.0 z: 1.0 }
    mass: 5.3155746
  }
  bodies {
    name: "floor"
    colliders {
      plane {}
    }
    inertia { x: 1.0 y: 1.0 z: 1.0 }
    frozen { all: true }
  }
  joints {
    name: "thigh_joint"
    parent: "torso"
    child: "thigh"
    parent_offset { z: -0.2 }
    rotation { z: -90.0 }
    angle_limit { min: -150.0 }
  }
  joints {
    name: "leg_joint"
    parent: "thigh"
    child: "leg"
    parent_offset { z: -0.45 }
    child_offset { z: 0.25 }
    rotation { z: -90.0 }
    angle_limit { min: -150.0 }
  }
  joints {
    name: "foot_joint"
    parent: "leg"
    child: "foot"
    parent_offset { z: -0.25 }
    child_offset {
      x: -0.13
      y: -0.13
      z: -0.03
    }
    rotation { z: -90.0 }
    angle_limit { min: -45.0 max: 45.0 }
  }
  actuators {
    name: "thigh_joint"
    joint: "thigh_joint"
    strength: 200.0
    torque {}
  }
  actuators {
    name: "leg_joint"
    joint: "leg_joint"
    strength: 200.0
    torque {}
  }
  actuators {
    name: "foot_joint"
    joint: "foot_joint"
    strength: 200.0
    torque {}
  }
  friction: 0.94868329805
  gravity { z: -9.81 }
  velocity_damping: 1.0
  angular_damping: -0.05
  collide_include {
    first: "floor"
    second: "torso"
  }
  collide_include {
    first: "floor"
    second: "thigh"
  }
  collide_include {
    first: "floor"
    second: "leg"
  }
  collide_include {
    first: "floor"
    second: "foot"
  }
  dt: 0.008
  substeps: 4
  frozen {
    position { y: 1.0 }
    rotation { x: 1.0 z: 1.0 }
  }
  defaults {
    qps { name: "torso" pos { z: 1.25 } }
    angles { name: "thigh_joint" angle {} }
    angles { name: "leg_joint" angle {} }
  }
  dynamics_mode: "pbd"
  """

_SYSTEM_CONFIG_SPRING = """
  bodies {
    name: "torso"
    colliders {
      capsule {
        radius: 0.05
        length: 0.5
      }
    }
    inertia { x: 1.0 y: 1.0 z: 1.0 }
    mass: 3.6651914
  }
  bodies {
    name: "thigh"
    colliders {
      position { z: -0.225 }
      rotation {}
      capsule {
        radius: 0.05
        length: 0.55
      }
    }
    inertia { x: 1.0 y: 1.0 z: 1.0 }
    mass: 4.0578904
  }
  bodies {
    name: "leg"
    colliders {
      position {}
      rotation {}
      capsule {
        radius: 0.04
        length: 0.58
      }
    }
    inertia { x: 1.0 y: 1.0 z: 1.0 }
    mass: 2.7813568
  }
  bodies {
    name: "foot"
    colliders {
      position {
        x: -0.065
        y: -0.13
        z: -0.03
      }
      rotation { y: 90.0 }
      capsule {
        radius: 0.06
        length: 0.51
      }
    }
    inertia { x: 1.0 y: 1.0 z: 1.0 }
    mass: 5.3155746
  }
  bodies {
    name: "floor"
    colliders {
      plane {}
    }
    inertia { x: 1.0 y: 1.0 z: 1.0 }
    frozen { all: true }
  }
  joints {
    name: "thigh_joint"
    stiffness: 10000.0
    parent: "torso"
    child: "thigh"
    parent_offset { z: -0.2 }
    rotation { z: -90.0 }
    angle_limit { min: -150.0 }
    angular_damping: 20.0
  }
  joints {
    name: "leg_joint"
    stiffness: 10000.0
    parent: "thigh"
    child: "leg"
    parent_offset { z: -0.45 }
    child_offset { z: 0.25 }
    rotation { z: -90.0 }
    angle_limit { min: -150.0 }
    angular_damping: 20.0
  }
  joints {
    name: "foot_joint"
    stiffness: 10000.0
    parent: "leg"
    child: "foot"
    parent_offset { z: -0.25 }
    child_offset {
      x: -0.13
      y: -0.13
      z: -0.03
    }
    rotation { z: -90.0 }
    angle_limit { min: -45.0 max: 45.0 }
    angular_damping: 20.0
  }
  actuators {
    name: "thigh_joint"
    joint: "thigh_joint"
    strength: 200.0
    torque {}
  }
  actuators {
    name: "leg_joint"
    joint: "leg_joint"
    strength: 200.0
    torque {}
  }
  actuators {
    name: "foot_joint"
    joint: "foot_joint"
    strength: 200.0
    torque {}
  }
  friction: 0.94868329805
  gravity { z: -9.81 }
  velocity_damping: 1.0
  angular_damping: -0.05
  baumgarte_erp: 0.1
  collide_include {
    first: "floor"
    second: "torso"
  }
  collide_include {
    first: "floor"
    second: "thigh"
  }
  collide_include {
    first: "floor"
    second: "leg"
  }
  collide_include {
    first: "floor"
    second: "foot"
  }
  dt: 0.008
  substeps: 2
  frozen {
    position { y: 1.0 }
    rotation { x: 1.0 z: 1.0 }
  }
  defaults {
    qps { name: "torso" pos { z: 1.25 } }
    angles { name: "thigh_joint" angle {} }
    angles { name: "leg_joint" angle {} }
  }
  dynamics_mode: "legacy_spring"
  """



# Copyright 2024 The Brax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Tests for brax.jumpy."""

from absl.testing import absltest
from brax.v1 import jumpy as jp
import jax
from jax import numpy as jnp
import numpy as np


class ForiLoopTest(absltest.TestCase):
  """Tests jumpy.fori_loop when jitted and not jitted."""

  def testForiLoopTest(self):
    a = jp.fori_loop(2, 4, lambda i, x: i + x, jp.array(1.))
    self.assertIsInstance(a, np.float_)
    self.assertEqual(a.shape, ())
    self.assertAlmostEqual(a, 1.0 + 2.0 + 3.0)

  def testForiLoopTestJit(self):
    a = jax.jit(lambda: jp.fori_loop(2, 4, lambda i, x: i + x, jp.array(1.)))()
    self.assertIsInstance(a, jnp.ndarray)
    self.assertEqual(a.shape, ())
    self.assertAlmostEqual(a, 1.0 + 2.0 + 3.0)


if __name__ == '__main__':
  absltest.main()



# Copyright 2024 The Brax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Tests for domain randomization."""

from absl.testing import absltest
from brax.v1.envs import ant
from brax.v1.experimental.tracing import randomizers
from brax.v1.experimental.tracing import wrappers
import brax.v1.jumpy as jp
import jax


class WrappersTest(absltest.TestCase):

  def test_randomize_ant_friction(self):

    test_ant_fn = ant.Ant

    # generate random friction values
    random_friction = jp.array(
        jax.random.uniform(jax.random.PRNGKey(42), (128,)))

    # build traceable config trees
    friction_tree, friction_axes = randomizers.friction_randomizer(
        test_ant_fn(), random_friction)

    random_friction_test_ant = wrappers.DomainRandomizationWrapper(
        test_ant_fn, friction_tree, friction_axes)

    # test reset
    out_state = jax.jit(random_friction_test_ant.reset)(
        jax.random.split(jax.random.PRNGKey(42), 128))
    self.assertEqual(out_state.qp.pos.shape[0], 128)

    # test step
    next_state = jax.jit(random_friction_test_ant.step)(out_state,
                                                        jp.zeros((128, 10)))
    self.assertEqual(next_state.qp.pos.shape[0], 128)


if __name__ == '__main__':
  absltest.main()
